import os
import logging
from mutagen.flac import FLAC
import mutagen.id3 as id3
from mutagen.id3 import ID3NoHeaderError

from qobuz_dl.lyrics_engine import LyricsEngine
from qobuz_dl.color import CYAN, GREEN, YELLOW, RED, OFF

logger = logging.getLogger(__name__)

def inject_lyrics_retroactively(directory_path, genius_token=None):
    print(f"\n{CYAN}[*] Starting retroactive lyrics scan in: {directory_path}{OFF}")
    
    if not os.path.isdir(directory_path):
        print(f"{RED}[!] Error: The directory '{directory_path}' does not exist.{OFF}")
        return

    engine = LyricsEngine(genius_token)
    processed = 0
    injected = 0
    skipped = 0
    
    for root, _, files in os.walk(directory_path):
        for file in files:
            if file.lower().endswith((".flac", ".mp3")):
                file_path = os.path.join(root, file)
                processed += 1
                
                try:
                    title, artist, album = "", "", ""
                    needs_lyrics = False

                    # --- FLAC HANDLING ---
                    if file.lower().endswith(".flac"):
                        audio = FLAC(file_path)
                        # Check if lyrics are already present
                        if audio.get("LYRICS") or audio.get("UNSYNCEDLYRICS"):
                            skipped += 1
                            continue
                            
                        title = audio.get("TITLE", [""])[0]
                        album_artist = audio.get("ALBUMARTIST", [""])[0]
                        performer_name = audio.get("ARTIST", ["Unknown Artist"])[0]
                        artist = performer_name if album_artist in ["", "Various Artists"] else album_artist
                        album = audio.get("ALBUM", [""])[0]
                        needs_lyrics = True
                        
                    # --- MP3 HANDLING ---
                    elif file.lower().endswith(".mp3"):
                        try:
                            audio = id3.ID3(file_path)
                        except ID3NoHeaderError:
                            skipped += 1
                            continue
                            
                        # Check if lyrics are already present (USLT = unsynced, SYLT = synced)
                        if audio.getall("USLT") or audio.getall("SYLT"):
                            skipped += 1
                            continue
                            
                        title = audio.get("TIT2").text[0] if audio.get("TIT2") else ""
                        artist = audio.get("TPE1").text[0] if audio.get("TPE1") else ""
                        album = audio.get("TALB").text[0] if audio.get("TALB") else ""
                        needs_lyrics = True

                    # Skip if file is corrupted or lacks basic tags
                    if not title or not artist:
                        skipped += 1
                        continue
                        
                    # --- LYRICS INJECTION ---
                    if needs_lyrics:
                        print(f"{YELLOW}  > Missing lyrics: {artist} - {title}. Searching...{OFF}")
                        
                        # Call the LyricsEngine
                        engine.fetch_and_inject(
                            file_path=file_path,
                            artist=artist,
                            track=title,
                            album=album
                        )
                        injected += 1
                            
                except Exception as e:
                    print(f"{RED}[!] Error reading {file}: {e}{OFF}")
                    skipped += 1
                    
    print(f"\n{GREEN}[+] Retroactive Scan and Injection Completed!{OFF}")
    print(f"{CYAN}  - Total files analyzed: {processed}{OFF}")
    print(f"{GREEN}  - Injection attempts: {injected}{OFF}")
    print(f"{YELLOW}  - Skipped files (already tagged or missing data): {skipped}{OFF}\n")


import argparse


def fun_args(subparsers, default_limit):
    interactive = subparsers.add_parser(
        "interactive",
        description="Interactively search for tracks and albums.",
        help="interactive mode",
        aliases=["i", "fun"]
    )
    interactive.add_argument(
        "-l",
        "--limit",
        metavar="int",
        default=default_limit,
        help="limit of search results (default: 20)",
    )
    return interactive


def lucky_args(subparsers):
    lucky = subparsers.add_parser(
        "lucky",
        description="Download the first <n> albums returned from a Qobuz search.",
        help="lucky mode",
    )
    lucky.add_argument(
        "-t",
        "--type",
        default="album",
        help="type of items to search (artist, album, track, playlist) (default: album)",
    )
    lucky.add_argument(
        "-n",
        "--number",
        metavar="int",
        default=1,
        help="number of results to download (default: 1)",
    )
    lucky.add_argument("QUERY", nargs="+", help="search query")
    return lucky


def dl_args(subparsers):
    download = subparsers.add_parser(
        "dl",
        description="Download by album/track/artist/label/playlist/last.fm-playlist URL.",
        help="input mode",
    )
    download.add_argument(
        "SOURCE",
        metavar="SOURCE",
        nargs="+",
        help=("one or more URLs (space separated) or a text file"),
    )
    
    # --- START BLACKLIST FLAG ---
    download.add_argument(
        "-b",
        "--blacklist",
        help="Path to a text file containing keywords to blacklist and skip",
        type=str,
        default=None,
    )
    # --- END BLACKLIST FLAG ---
    
    return download

def lyrics_args(subparsers):
    lyrics = subparsers.add_parser(
        "lyrics",
        description="Retroactively scan a directory and inject missing lyrics into existing audio files.",
        help="lyrics injection mode",
    )
    lyrics.add_argument(
        "DIR",
        metavar="DIRECTORY",
        help="The local directory containing the music files to be scanned",
    )
    return lyrics

def sync_playlist_args(subparsers):
    sync_pl = subparsers.add_parser(
        "sync-playlist",
        aliases=["sp"],
        description="Synchronize a local folder with a Qobuz playlist. "
                    "Downloads missing tracks and removes tracks no longer in the playlist.",
        help="sync a local folder with a Qobuz playlist",
    )
    sync_pl.add_argument(
        "URL",
        help="Qobuz playlist URL (e.g. https://play.qobuz.com/playlist/12345)",
    )
    sync_pl.add_argument(
        "--yes", "-y",
        action="store_true",
        help="Skip confirmation prompt before deleting/downloading",
    )
    return sync_pl

def add_common_arg(custom_parser, default_folder, default_quality):
    custom_parser.add_argument(
        "-d",
        "--directory",
        metavar="PATH",
        default=default_folder,
        help=f'directory for downloads (default: "{default_folder}")',
    )
    custom_parser.add_argument(
        "--no-lrc-files",
        dest="lrc_files",
        action="store_false",
        default=argparse.SUPPRESS,
        help="do not save synchronized lyrics to external .lrc files",
    )
    custom_parser.add_argument(
        "-q",
        "--quality",
        metavar="int",
        type=int,
        default=default_quality,
        choices=[5, 6, 7, 27],
        help=(
            'audio "quality" (5, 6, 7, 27)\n'
            f"[320, LOSSLESS, 24B<=96KHZ, 24B>96KHZ] (default: {default_quality})"
        ),
    )
    custom_parser.add_argument(
        "--albums-only",
        action="store_true",
        help=("don't download singles, EPs and VA releases"),
    )
    custom_parser.add_argument(
        "--no-m3u",
        action="store_true",
        help="don't create .m3u files when downloading playlists",
    )
    custom_parser.add_argument(
        "--no-fallback",
        action="store_true",
        help="disable quality fallback (skip releases not available in set quality)",
    )
    custom_parser.add_argument(
        "--no-db", action="store_true", help="don't call the database"
    )
    custom_parser.add_argument(
        "-ff",
        "--folder-format",
        metavar="PATTERN",
        help="""pattern for formatting folder names, e.g
        "{album_artist} - {album_title} ({year}) {{{barcode}}}". available keys: 
        album_id, album_url, album_title, album_title, album_artist, album_genre, 
        album_composer, label, copyright, upc, barcode, release_date, year, media_type,
        format, bit_depth, sampling_rate, album_version, disc_count, track_count.
        Note1: {album_title}, {track_title} will contain version information if available.
        Note2: {album_title_base}, {track_title_base} will contain only the title,
        Note3: {track_title}, {track_title_base} is only available if the given url is a track url.
        Note4: You can use '/' to create subdirectories, for example:
        "{album_artist}/{album_artist} - {album_title} ({year})" will create
        "Taylor Swift/Taylor Swift - folklore (2020)".
        Cannot contain characters used by the system, which includes :<>""",
    )
    custom_parser.add_argument(
        "-fbff",
        "--fallback-folder-format",
        metavar="PATTERN", 
        help="""fallback pattern for formatting folder names when the main pattern fails.
        Uses same keys as --folder-format. e.g: "{album_artist} - {album_title}"
        Note: You can also use '/' to create subdirectories in the fallback pattern.""",
    )
    custom_parser.add_argument(
        "-tf",
        "--track-format",
        metavar="PATTERN",
        help="""pattern for formatting track names. e.g
        "{track_number} - {track_title}" 
        available keys:
        album_title, album_title_base, album_artist, track_id, track_artist, track_composer, 
        track_number, isrc, bit_depth, sampling_rate, track_title, track_title_base
        version, year, disc_number, release_date.
        Note1: {album_title}, {track_title} will contain version information if available.
        Note2: {album_title_base}, {track_title_base} will contain only the title.
        Cannot contain characters used by the system, which includes /:<>
        """,
    )
    custom_parser.add_argument(
        "-s",
        "--smart-discography",
        action="store_true",
        help="""Try to filter out spam-like albums when requesting an artist's
        discography, and other optimizations. Filters albums not made by requested
        artist, and deluxe/live/collection albums. Gives preference to remastered
        albums, high bit depth/dynamic range, and low sampling rates (to save space).""",
    )
    
    # --- HUMAN BEHAVIOR DELAY ---
    custom_parser.add_argument(
        "--delay",
        type=int,
        default=0,
        help="Wait a specified number of seconds between track downloads to prevent server bans.",
    )

    # --- NEW COMMANDS FOR ULTIMATE FEATURES ---
    custom_parser.add_argument(
        "--no-lyrics",
        action="store_true",
        help="disable automatic lyrics fetching and injection for this session",
    )
    custom_parser.add_argument(
        "--no-embed-lyrics",
        dest="no_embed_lyrics",
        action="store_true",
        default=argparse.SUPPRESS,
        help="do not embed lyrics into the audio file tags (save as .lrc/.txt only)",
    )
    custom_parser.add_argument(
        "--multi-tags",
        dest="multi_value_tags",
        action="store_true",
        default=argparse.SUPPRESS,
        help="split comma-separated metadata (like genres) into multiple tag fields",
    )
    custom_parser.add_argument(
        "--booklet-only",
        action="store_true",
        help="only download the Digital Booklet and PDF Goodies (skips audio files)",
    )
    custom_parser.add_argument(
        "--native-lang",
        action="store_true",
        help="do not force English; download metadata in the account's native language",
    )
    custom_parser.add_argument(
        "--no-credits",
        action="store_true",
        help="disable the generation of the Digital Booklet.txt (Credits & Review) file",
    )

    # Add override flag to force credits generation if no_credits is set to true in config.ini
    custom_parser.add_argument(
        "--with-credits",
        action="store_true",
        help="force the generation of the Digital Booklet.txt (overrides config.ini)",
    )

    # Adding tag-related parameters
    tag_group = custom_parser.add_argument_group('tag options')
    tag_group.add_argument(
        "--no-album-artist-tag", 
        action="store_true",
        help="don't add album artist tag"
    )
    tag_group.add_argument(
        "--no-album-title-tag",
        action="store_true", 
        help="don't add album title tag"
    )
    tag_group.add_argument(
        "--no-track-artist-tag",
        action="store_true",
        help="don't add track artist tag"
    )
    tag_group.add_argument(
        "--no-track-title-tag",
        action="store_true",
        help="don't add track title tag"
    )
    tag_group.add_argument(
        "--no-release-date-tag",
        action="store_true",
        help="don't add release date tag"
    )
    tag_group.add_argument(
        "--no-media-type-tag",
        action="store_true",
        help="don't add media type tag"
    )
    tag_group.add_argument(
        "--no-genre-tag",
        action="store_true",
        help="don't add genre tag"
    )
    tag_group.add_argument(
        "--no-replaygain-tag",
        action="store_true",
        help="Do not add ReplayGain tags to the audio files.",
    )
    tag_group.add_argument(
        "--no-album-url-tag",
        action="store_true",
        help="Do not add the QOBUZ ALBUM URL tag to the audio files.",
    )
    tag_group.add_argument(
        "--no-track-number-tag",
        action="store_true",
        help="don't add track number tag"
    )
    tag_group.add_argument(
        "--no-track-total-tag",
        action="store_true",
        help="don't add total tracks tag"
    )
    tag_group.add_argument(
        "--no-disc-number-tag",
        action="store_true",
        help="don't add disc number tag"
    )
    tag_group.add_argument(
        "--no-disc-total-tag",
        action="store_true",
        help="don't add total discs tag"
    )
    tag_group.add_argument(
        "--no-composer-tag",
        action="store_true",
        help="don't add composer tag"
    )
    tag_group.add_argument(
        "--no-explicit-tag",
        action="store_true",
        help="don't add explicit advisory tag"
    )
    tag_group.add_argument(
        "--no-copyright-tag",
        action="store_true",
        help="don't add copyright tag"
    )
    tag_group.add_argument(
        "--no-label-tag",
        action="store_true",
        help="don't add label tag"
    )
    tag_group.add_argument(
        "--no-upc-tag",
        action="store_true",
        help="don't add UPC/barcode tag"
    )
    tag_group.add_argument(
        "--no-isrc-tag",
        action="store_true",
        help="don't add ISRC tag"
    )
    artwork_group = custom_parser.add_argument_group('cover artwork options')
    artwork_group.add_argument(
        "-e", "--embed-art", action="store_true", help="embed cover art into audio files"
    )
    artwork_group.add_argument(
        "--og-cover",
        action="store_true",
        help="download cover art in its original quality (bigger file). No longer available, recommended use: --embedded-art-size and --saved-art-size",
    )
    artwork_group.add_argument(
        "--no-cover", action="store_true", help="don't download cover art"
    )
    artwork_group.add_argument(
        "--embedded-art-size",
        choices=["50", "100", "150", "300", "600", "max", "org"],
        default="600",
        help="size of embedded artwork (default: 600)"
    )
    artwork_group.add_argument(
        "--saved-art-size",
        choices=["50", "100", "150", "300", "600", "max", "org"],
        default="org",
        help="size of saved artwork (default: org)"
    )
    multiple_disc_group = custom_parser.add_argument_group('multiple disc options')
    multiple_disc_group.add_argument(
        "--multiple-disc-prefix",
        default="CD",
        metavar="PREFIX",
        help="""Setting folder prefix for multiple discs album (default: CD)
        If the album has multiple discs(media_count > 1), the album's tracks will be saved by folder.
        The names of the folders: '{prefix} {media_number}', eg: 'CD 01'
        """
    )
    multiple_disc_group.add_argument(
        "--multiple-disc-one-dir",
        action="store_true",
        help="store multiple disc releases in one directory",
    )
    multiple_disc_group.add_argument(
        "--multiple-disc-track-format",
        metavar="FORMAT",
        help='track format for multiple disc releases (default: "{disc_number}.{track_number} - {track_title}")',
    )

    # Add parallel download thread count argument group
    parallel_group = custom_parser.add_argument_group('parallel download options')
    parallel_group.add_argument(
        "--max-workers",
        type=int,
        metavar="N",
        help="maximum number of parallel downloads (default: 3)",
    )


def qobuz_dl_args(
    default_quality=6, default_limit=20, default_folder="QobuzDownloads"
):
    parser = argparse.ArgumentParser(
        prog="qobuz-dl",
        description=(
            "The ultimate Qobuz music downloader.\nSee usage"
            " examples on https://github.com/Sei969/qobuz-dl"
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-r", "--reset", action="store_true", help="create/reset config file"
    )
    parser.add_argument(
        "-p",
        "--purge",
        action="store_true",
        help="purge/delete downloaded-IDs database",
    )
    parser.add_argument(
        "--sync-db",
        metavar="PATH",
        nargs="?",
        const="DEFAULT",
        help="scan local directory to restore missing Qobuz IDs into the database",
    )
    parser.add_argument(
        "-sc",
        "--show-config",
        action="store_true",
        help="show configuration",
    )

    subparsers = parser.add_subparsers(
        title="commands",
        description="run qobuz-dl <command> --help for more info\n(e.g. qobuz-dl interactive --help)",
        dest="command",
    )

    interactive = fun_args(subparsers, default_limit)
    download = dl_args(subparsers)
    lucky = lucky_args(subparsers)
    
    # Inizializza il nuovo comando
    lyrics_cmd = lyrics_args(subparsers)
    sync_pl_cmd = sync_playlist_args(subparsers)
    
    [
        add_common_arg(i, default_folder, default_quality)
        for i in (interactive, download, lucky, sync_pl_cmd)
    ]

    return parser


import sys
import difflib
import string
import configparser
import logging
import glob
import os
import getpass
import hashlib
import signal
import keyring
import requests

from qobuz_dl.bundle import Bundle
from qobuz_dl.color import GREEN, RED, YELLOW, OFF, CYAN
from qobuz_dl.commands import qobuz_dl_args
from qobuz_dl.core import QobuzDL
from qobuz_dl.downloader import DEFAULT_FOLDER, DEFAULT_TRACK
from qobuz_dl.settings import QobuzDLSettings

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
)

if os.name == "nt":
    OS_CONFIG = os.environ.get("APPDATA")
else:
    OS_CONFIG = os.path.join(os.environ["HOME"], ".config")

CONFIG_PATH = os.path.join(OS_CONFIG, "qobuz-dl")
CONFIG_FILE = os.path.join(CONFIG_PATH, "config.ini")
QOBUZ_DB = os.path.join(CONFIG_PATH, "qobuz_dl.db")

KEYRING_SERVICE = "qobuz-dl"


def _keyring_save(key, value):
    if not value:
        return False
    try:
        keyring.set_password(KEYRING_SERVICE, key, value)
        return True
    except Exception:
        return False


def _keyring_load(key):
    try:
        return keyring.get_password(KEYRING_SERVICE, key)
    except Exception:
        return None


def validate_config_formats(formats_to_check):
    """
    Scans the configuration format strings for unknown variables to prevent
    silent KeyErrors during the download process. Includes typo suggestions.
    """
    VALID_KEYS = {
        "artist", "album", "album_id", "album_url", "album_title", 
        "album_title_base", "album_artist", "album_genre", "album_composer", 
        "label", "copyright", "upc", "barcode", "release_date", "year", 
        "media_type", "format", "bit_depth", "sampling_rate", "album_version", 
        "version_tag", "disc_count", "track_count", "ExplicitFlag", "explicit", 
        "release_type", "tracktitle", "track_title", "track_title_base", 
        "track_id", "track_artist", "track_composer", "track_number", 
        "isrc", "version", "disc_number"
    }

    has_errors = False
    
    # Define color strings locally to ensure they print correctly in terminal
    C_RED = '\033[91m'
    C_YEL = '\033[93m'
    C_GRE = '\033[92m'
    C_OFF = '\033[0m'

    for config_name, format_string in formats_to_check.items():
        if not format_string:
            continue
            
        try:
            parsed_vars = [tup[1] for tup in string.Formatter().parse(str(format_string)) if tup[1] is not None]
            
            for var in parsed_vars:
                base_var = var.split(':')[0].split('!')[0]
                
                if base_var not in VALID_KEYS:
                    print(f"{C_YEL}[!] Config Warning: Unknown variable '{{{base_var}}}' detected in '{config_name}'.{C_OFF}")
                    
                    # --- NEW: 'Did you mean' logic using difflib ---
                    similar_keys = difflib.get_close_matches(base_var, VALID_KEYS, n=1, cutoff=0.6)
                    if similar_keys:
                        print(f"    {C_GRE}-> Did you mean '{{{similar_keys[0]}}}'?{C_OFF}")
                    # -----------------------------------------------
                    
                    print(f"    {C_RED}-> This will cause the entire format string to be discarded during download.{C_OFF}")
                    has_errors = True
                    
        except ValueError as e:
            print(f"{C_RED}[!] Config Error: Syntax error in '{config_name}' -> {e}{C_OFF}")
            has_errors = True

    if has_errors:
        print(f"\n{C_YEL}[*] Tip: Please check your config.ini file or your command line arguments and fix any typos before downloading.{C_OFF}\n")
        # Abort the process immediately
        sys.exit(1)


def _reset_config(config_file):
    logging.info(f"\n{YELLOW}--- QOBUZ-DL CONFIGURATION WIZARD (2026 Update) ---{OFF}")
    config = configparser.ConfigParser(interpolation=None)
    
    config["qobuz"] = {}
    
    email = input("Enter your Qobuz email:\n- ").strip()
    config["qobuz"]["email"] = email
    
    print(f"\n{YELLOW}[!] ATTENTION: Qobuz API blocked direct password login for 3rd party apps.{OFF}")
    print(f"{YELLOW}[!] You must use your browser Auth Token (F12 > Storage > Local Storage > localuser > token).{OFF}")
    
    auth_token = input("Paste your browser token here:\n- ").strip()

    config["qobuz"]["password"] = ""
    config["qobuz"]["auth_token"] = "" if _keyring_save("auth_token", auth_token) else auth_token

    fetch_lyrics = input("Do you want to automatically download and inject lyrics? (yes/no) [Default: yes]\n- ").strip().lower()
    config["qobuz"]["fetch_lyrics"] = "false" if fetch_lyrics in ['no', 'n', 'false'] else "true"
    
    genius_token = ""
    if config["qobuz"]["fetch_lyrics"] == "true":
        print(f"{YELLOW}[!] To use Genius as a fallback, enter your API Token. Leave blank to only use LRCLIB (Free/No API).{OFF}")
        genius_token = input("Genius API Token:\n- ").strip()
    config["qobuz"]["genius_token"] = "" if _keyring_save("genius_token", genius_token) else genius_token

    config["qobuz"]["directory"] = (
        input("Download folder (press Enter for 'Qobuz Downloads')\n- ")
        or "Qobuz Downloads"
    )
    
    # FIX: Use correct prompt and key for folder formatting
    config["qobuz"]["folder_format"] = (
        input(f"Folder format (press Enter for '{DEFAULT_FOLDER}')\n- ")
        or DEFAULT_FOLDER
    )
    
    config["qobuz"]["default_quality"] = (
        input("Download quality (5:MP3, 6:FLAC, 7:24b<96, 27:24b>96) [Default 27]\n- ")
        or "27"
    )
    
    config["qobuz"]["default_limit"] = "500"
    config["qobuz"]["no_m3u"] = "false"
    config["qobuz"]["albums_only"] = "false"
    config["qobuz"]["no_fallback"] = "false"
    config["qobuz"]["og_cover"] = "true" 
    config["qobuz"]["embed_art"] = "true"
    config["qobuz"]["no_cover"] = "false"
    config["qobuz"]["no_database"] = "false"
    config["qobuz"]["no_lrc_files"] = "true"
    config["qobuz"]["embed_lyrics"] = "true"
    config["qobuz"]["multi_value_tags"] = "false"
    config["qobuz"]["legacy_charmap"] = "false"
    config["qobuz"]["blacklist"] = "blacklist.txt"

    logging.info(f"{YELLOW}Getting tokens. Please wait...{OFF}")
    bundle = Bundle()
    config["qobuz"]["app_id"] = str(bundle.get_app_id())
    config["qobuz"]["secrets"] = ",".join(bundle.get_secrets().values())

    # Removed old folder_format override that caused custom format resets
    config["qobuz"]["track_format"] = "{track_number} - {track_title}"
    config["qobuz"]["fallback_folder_format"] = "{artist} - {album}"
    config["qobuz"]["smart_discography"] = "false"

    config["qobuz"]["no_album_artist_tag"] = "false"
    config["qobuz"]["no_album_title_tag"] = "false"
    config["qobuz"]["no_track_artist_tag"] = "false"
    config["qobuz"]["no_track_title_tag"] = "false"
    config["qobuz"]["no_release_date_tag"] = "false"
    config["qobuz"]["no_media_type_tag"] = "false"
    config["qobuz"]["no_genre_tag"] = "false"
    config["qobuz"]["no_track_number_tag"] = "false"
    config["qobuz"]["no_track_total_tag"] = "false"
    config["qobuz"]["no_disc_number_tag"] = "false"
    config["qobuz"]["no_disc_total_tag"] = "false"
    config["qobuz"]["no_composer_tag"] = "false"
    config["qobuz"]["no_replaygain_tag"] = "false"
    config["qobuz"]["no_album_url_tag"] = "false"
    
    config["qobuz"]["no_explicit_tag"] = "false"
    config["qobuz"]["no_copyright_tag"] = "false"
    config["qobuz"]["no_label_tag"] = "false"
    
    config["qobuz"]["no_credits"] = "false"
    
    config["qobuz"]["no_upc_tag"] = "false"
    config["qobuz"]["no_isrc_tag"] = "false"
          
    config["qobuz"]["embedded_art_size"] = "600"
    config["qobuz"]["saved_art_size"] = "org"
    
    config["qobuz"]["multiple_disc_prefix"] = "CD"
    config["qobuz"]["multiple_disc_one_dir"] = "false"
    config["qobuz"]["multiple_disc_track_format"] = "{disc_number}.{track_number} - {track_title}"
    
    config["qobuz"]["max_workers"] = "3"
    config["qobuz"]["user_auth_token"] = ""
    
    with open(config_file, "w") as configfile:
        config.write(configfile)
        
    logging.info(f"\n{GREEN}[+] Configuration successfully saved in {config_file}!{OFF}")
    

def _remove_leftovers(directory):
    for pattern in [".*.tmp", "~tmp_*.tmp"]:
        search_dir = os.path.join(directory, "**", pattern)
        for i in glob.glob(search_dir, recursive=True):
            try:
                os.remove(i)
            except:  # noqa
                pass


def _handle_commands(qobuz, arguments):
    def sigint_handler(sig, frame):
        print(f"\n\n\033[91m[!] Download forcibly interrupted by the user.\033[0m")
        print(f"\033[93mPartially downloaded files will be ignored or overwritten on the next run.\033[0m")
        try:
            _remove_leftovers(qobuz.directory)
        except Exception:
            pass
        sys.exit(1)
        
    signal.signal(signal.SIGINT, sigint_handler)

    try:
        if arguments.command == "dl":
            qobuz.download_list_of_urls(arguments.SOURCE)
        elif arguments.command in ("sync-playlist", "sp"):
            from qobuz_dl.sync_playlist import sync_playlist
            sync_playlist(
                qobuz,
                arguments.URL,
                qobuz.directory,  # <-- MODIFIED: Previously it was arguments.FOLDER
                auto_confirm=arguments.yes,
            )
        elif arguments.command == "lucky":
            query = " ".join(arguments.QUERY)
            qobuz.lucky_type = arguments.type
            qobuz.lucky_limit = arguments.number
            qobuz.lucky_mode(query)
        else:
            qobuz.interactive_limit = arguments.limit
            qobuz.interactive()

    except KeyboardInterrupt:
        pass
    finally:
        _remove_leftovers(qobuz.directory)


def _initial_checks():
    if not os.path.isdir(CONFIG_PATH) or not os.path.isfile(CONFIG_FILE):
        os.makedirs(CONFIG_PATH, exist_ok=True)
        if "-r" not in sys.argv and "--reset" not in sys.argv:
            _reset_config(CONFIG_FILE)

    if len(sys.argv) < 2:
        sys.exit(qobuz_dl_args().print_help())

def check_for_updates():
    try:
        from qobuz_dl import __version__
        
        url = "https://api.github.com/repos/Sei969/qobuz-dl/releases/latest"
        response = requests.get(url, timeout=2)
        response.raise_for_status()
        
        latest_version_str = response.json().get("tag_name", "").replace("v", "")
        current_version_str = __version__
        
        latest_tuple = tuple(map(int, latest_version_str.split(".")))
        current_tuple = tuple(map(int, current_version_str.split(".")))
        
        if latest_tuple > current_tuple:
            print(f"\n{YELLOW}[*] UPDATE AVAILABLE: Ultimate Edition v{latest_version_str} is out!{OFF}")
            print(f"{YELLOW}    - PyPI: run 'pip install -U qobuz-dl-ultimate'{OFF}")
            print(f"{YELLOW}    - Docker: pull the latest image{OFF}")
            print(f"{YELLOW}    - Standalone: download the new release from GitHub{OFF}\n")
            
    except Exception:
        pass

def main():
    _initial_checks()
    check_for_updates()

    # --- RADAR FEATURE (Standalone Intercept) ---
    import sys
    if len(sys.argv) > 1 and sys.argv[1] == "radar":
        from qobuz_dl.radar import run_radar
        
        try:
            run_radar()
        except KeyboardInterrupt:
            print("\n\n\033[91m[!] Radar manually interrupted by the user (CTRL+C).\033[0m")
        sys.exit(0)
    # --------------------------------------------

    # --- NEW: STATS COMMAND INTEGRATION ---
    if len(sys.argv) > 1 and sys.argv[1] == "stats":
        from qobuz_dl.db import get_stats
        
        # QOBUZ_DB è già definito all'inizio di cli.py, lo usiamo direttamente
        artists = get_stats(QOBUZ_DB)
        
        print(f"\n{CYAN}--- QOBUZ-DL ULTIMATE STATISTICS ---{OFF}")
        if not artists:
            print(f"{YELLOW}No artist data found yet. Start downloading to populate your stats!{OFF}")
        else:
            print(f"Total Unique Artists Downloaded: {len(artists)}\n")
            for artist in artists:
                print(f" - {artist}")
        print(f"{CYAN}-------------------------------------{OFF}\n")
        sys.exit(0) # Esce immediatamente dopo aver stampato le statistiche
    # -------------------------------------------------

    config = configparser.ConfigParser(interpolation=None)
    config.read(CONFIG_FILE)

    try:
        section = "qobuz" if config.has_section("qobuz") else "DEFAULT"
        
        email = config.get(section, "email")
        ini_token = config.get(section, "auth_token", fallback="")
        token = _keyring_load("auth_token") or ini_token
        password = token if token else config.get(section, "password")

        fetch_lyrics = config.getboolean(section, "fetch_lyrics", fallback=False)
        ini_genius = config.get(section, "genius_token", fallback=None)
        genius_token = _keyring_load("genius_token") or ini_genius

        migrated = False
        for k, v in (("auth_token", ini_token), ("genius_token", ini_genius)):
            if v and _keyring_save(k, v):
                config.set(section, k, "")
                migrated = True
        if migrated:
            try:
                with open(CONFIG_FILE, "w") as f:
                    config.write(f)
            except OSError:
                pass
        
        # --- FIX: Backward compatibility for default_folder ---
        directory_val = config.get(section, "directory", fallback=None)
        if directory_val is not None:
            default_folder = directory_val
        else:
            legacy_val = config.get(section, "default_folder", fallback=None)
            if legacy_val is not None:
                # If the legacy key is used, accept it but print a yellow warning
                print(f"\033[93m[!] Notice: 'default_folder' in config.ini is deprecated. Please rename it to 'directory' for future updates.\033[0m")
                default_folder = legacy_val
            else:
                default_folder = "Qobuz Downloads"
        # ------------------------------------------------------
        default_limit = config.get(section, "default_limit")
        default_quality = config.get(section, "default_quality")
        
        no_m3u = config.getboolean(section, "no_m3u", fallback=False)
        no_lrc_files_config = config.getboolean(section, "no_lrc_files", fallback=False)
        albums_only = config.getboolean(section, "albums_only", fallback=False)
        no_fallback = config.getboolean(section, "no_fallback", fallback=False)
        og_cover = config.getboolean(section, "og_cover", fallback=True)
        embed_art = config.getboolean(section, "embed_art", fallback=True)
        no_cover = config.getboolean(section, "no_cover", fallback=False)
        no_database = config.getboolean(section, "no_database", fallback=False)
        legacy_charmap = config.getboolean(section, "legacy_charmap", fallback=False)
        
        no_credits_config = config.getboolean(section, "no_credits", fallback=False)
        blacklist_config = config.get(section, "blacklist", fallback="blacklist.txt")
        
        app_id = config.get(section, "app_id")
        secrets = [s for s in config.get(section, "secrets").split(",") if s]
        
        smart_discography = config.getboolean(section, "smart_discography", fallback=False)
        folder_format = config.get(section, "folder_format", fallback=DEFAULT_FOLDER)
        track_format = config.get(section, "track_format", fallback=DEFAULT_TRACK)

        arguments = qobuz_dl_args(
            default_quality, default_limit, default_folder
        ).parse_args()
        
        if getattr(arguments, 'no_lyrics', False):
            fetch_lyrics = False
            
        force_english = not getattr(arguments, 'native_lang', False)
        no_credits_flag = getattr(arguments, 'no_credits', False) or no_credits_config 
        
    except (configparser.Error, KeyError) as error:
        arguments = qobuz_dl_args().parse_args()
        if not arguments.reset:
            # FIX: Define ANSI codes locally to bypass UnboundLocalError
            RED_C = '\033[91m'
            YELLOW_C = '\033[93m'
            OFF_C = '\033[0m'
            sys.exit(
                f"{RED_C}Invalid or corrupted configuration ({error}).\n{OFF_C}"
                f"{YELLOW_C}Run 'python -m qobuz_dl -r' to fix this.{OFF_C}"
            )

    if arguments.reset:
        sys.exit(_reset_config(CONFIG_FILE))

    if arguments.show_config:
        print(f"Configuration: {CONFIG_FILE}\nDatabase: {QOBUZ_DB}\n---")
        with open(CONFIG_FILE, "r") as f:
            print(f.read())
        sys.exit()

    if arguments.purge:
        try:
            os.remove(QOBUZ_DB)
        except FileNotFoundError:
            pass
        sys.exit(f"{GREEN}Database has been purged.{OFF}")

    # --- NEW DB SYNC FEATURE (Lightweight Mode) ---
    if getattr(arguments, 'sync_db', None):
        from qobuz_dl.sync import sync_database
        from qobuz_dl.qopy import Client
                
        # Initialize a lightweight API client for Reverse Lookup (bypassing the heavy downloader)
        sync_client = Client(email, password, app_id, secrets, user_auth_token=token, force_english=force_english)
        
        # Path management
        sync_dir = default_folder if arguments.sync_db == "DEFAULT" else arguments.sync_db
        
        if os.name == "nt":
            sync_dir = os.path.abspath(sync_dir)
            if not sync_dir.startswith("\\\\?\\"):
                sync_dir = "\\\\?\\" + sync_dir
                
        sync_database(sync_dir, QOBUZ_DB, sync_client)
        sys.exit(f"\n{GREEN}Database synchronization finished successfully.{OFF}")
    # ----------------------------------------------

    # --- RETRO LYRICS FEATURE (Standalone Mode) ---
    # Intercept the command here before QobuzDLSettings looks for 'directory', which would crash the program
    if arguments.command == "lyrics":
        from qobuz_dl.retro_tagger import inject_lyrics_retroactively
        
        target_dir = arguments.DIR
        if os.name == "nt":
            target_dir = os.path.abspath(target_dir)
            if not target_dir.startswith("\\\\?\\"):
                target_dir = "\\\\?\\" + target_dir
                
        try:
            inject_lyrics_retroactively(target_dir, genius_token=genius_token)
        except KeyboardInterrupt:
            print("\n\n\033[91m[!] Operation manually interrupted by the user (CTRL+C).\033[0m")
            print("\033[93mAlready processed files are safe. Exiting...\033[0m")
        sys.exit(0)
    # ----------------------------------------------

    directory_to_use = arguments.directory if hasattr(arguments, 'directory') and arguments.directory else default_folder
    directory_to_use = os.path.expanduser(directory_to_use)

    # --- WINDOWS LONG PATH BYPASS ---
    if os.name == "nt":
        directory_to_use = os.path.abspath(directory_to_use)
        if not directory_to_use.startswith("\\\\?\\"):
            directory_to_use = "\\\\?\\" + directory_to_use
    # --------------------------------

    settings = QobuzDLSettings.from_arguments_configparser(arguments, config)
    settings.legacy_charmap = legacy_charmap
    
    # Execute the Pre-flight Config Check
    # --- PRE-FLIGHT CONFIG CHECK ---
    formats_to_validate = {
        "folder_format": arguments.folder_format or folder_format,
        "track_format": arguments.track_format or track_format,
        "fallback_folder_format": config.get(section, "fallback_folder_format", fallback="{artist} - {album}"),
        "multiple_disc_track_format": config.get(section, "multiple_disc_track_format", fallback="{disc_number}.{track_number} - {track_title}")
    }
    validate_config_formats(formats_to_validate)
    # -------------------------------

    qobuz = QobuzDL(
        directory_to_use,
        arguments.quality,
        arguments.embed_art or embed_art,
        ignore_singles_eps=arguments.albums_only or albums_only,
        no_m3u_for_playlists=arguments.no_m3u or no_m3u,
        quality_fallback=not arguments.no_fallback or not no_fallback,
        cover_og_quality=arguments.og_cover or og_cover,
        no_cover=arguments.no_cover or no_cover,
        downloads_db=None if no_database or arguments.no_db else QOBUZ_DB,
        folder_format=arguments.folder_format or folder_format,
        track_format=arguments.track_format or track_format,
        smart_discography=arguments.smart_discography or smart_discography,
        fetch_lyrics=fetch_lyrics,
        no_lrc_files=("--no-lrc-files" in sys.argv) or no_lrc_files_config,
        genius_token=genius_token,
        force_english=force_english,
        no_credits=no_credits_flag,
        settings=settings,
        booklet_only=getattr(arguments, 'booklet_only', False),
        blacklist=getattr(arguments, 'blacklist', None) or blacklist_config,
    )
    
    qobuz.initialize_client(email, password, app_id, secrets)

    _handle_commands(qobuz, arguments)


if __name__ == "__main__":
    main()


import logging
import sqlite3

from qobuz_dl.color import YELLOW, RED, OFF

logger = logging.getLogger(__name__)


def create_db(db_path):
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()
        
        # Check if the table already exists
        cursor.execute("SELECT count(name) FROM sqlite_master WHERE type='table' AND name='downloads'")
        
        if cursor.fetchone()[0] == 1:
            # Table exists. Read current columns
            cursor.execute("PRAGMA table_info(downloads)")
            columns = [info[1] for info in cursor.fetchall()]
            
            # Legacy migration (v1 to v2)
            if 'quality' not in columns:
                logger.info(f"{YELLOW}Migrating old database to the new format...{OFF}")
                
                # Rename the old table
                conn.execute("ALTER TABLE downloads RENAME TO downloads_old")
                
                # Create the new table with updated schema including artist and album
                conn.execute("""
                CREATE TABLE downloads (
                  "id" text NOT NULL,
                  "media_type" text NOT NULL DEFAULT 'album',
                  "quality" integer NOT NULL DEFAULT 27,
                  "file_format" text NOT NULL DEFAULT 'FLAC',
                  "quality_met" integer NOT NULL DEFAULT 0,
                  "bit_depth" text,
                  "sampling_rate" text,
                  "saved_path" text NOT NULL DEFAULT '',
                  "status" text NOT NULL DEFAULT 'downloaded',
                  "url" text NOT NULL DEFAULT '',
                  "release_date" text NOT NULL DEFAULT '',
                  "artist" text NOT NULL DEFAULT '',
                  "album" text NOT NULL DEFAULT '',
                  PRIMARY KEY ("id", "quality")
                );
                """)
                
                # Copy old historical IDs
                try:
                    conn.execute("INSERT INTO downloads (id) SELECT id FROM downloads_old")
                except sqlite3.Error as e:
                    logger.error(f"{RED}Failed to migrate old data: {e}{OFF}")
                
                # Drop the temporary old table
                conn.execute("DROP TABLE downloads_old")
                logger.info(f"{YELLOW}Database successfully updated!{OFF}")
                
            # New Migration (v2 to v2.1.4): Add artist and album if missing
            elif 'artist' not in columns:
                logger.info(f"{YELLOW}Upgrading database schema: Adding artist and album columns...{OFF}")
                try:
                    conn.execute("ALTER TABLE downloads ADD COLUMN artist text NOT NULL DEFAULT ''")
                    conn.execute("ALTER TABLE downloads ADD COLUMN album text NOT NULL DEFAULT ''")
                    logger.info(f"{YELLOW}Schema upgrade complete!{OFF}")
                except sqlite3.Error as e:
                    logger.error(f"{RED}Failed to add new columns: {e}{OFF}")
                
        else:
            # Table does not exist, create it from scratch
            try:
                conn.execute("""
                CREATE TABLE downloads (
                  "id" text NOT NULL,
                  "media_type" text NOT NULL DEFAULT 'album',
                  "quality" integer NOT NULL DEFAULT 27,
                  "file_format" text NOT NULL DEFAULT 'FLAC',
                  "quality_met" integer NOT NULL DEFAULT 0,
                  "bit_depth" text,
                  "sampling_rate" text,
                  "saved_path" text NOT NULL DEFAULT '',
                  "status" text NOT NULL DEFAULT 'downloaded',
                  "url" text NOT NULL DEFAULT '',
                  "release_date" text NOT NULL DEFAULT '',
                  "artist" text NOT NULL DEFAULT '',
                  "album" text NOT NULL DEFAULT '',
                  PRIMARY KEY ("id", "quality")
                );
                """)
                logger.info(f"{YELLOW}Download-IDs database created{OFF}")
            except sqlite3.OperationalError:
                pass
                
        return db_path


def handle_download_id(db_path, item_id, add_id=False, media_type='album', quality=27, file_format='FLAC',
                       quality_met=0, bit_depth=None, sampling_rate=None, saved_path='', status='downloaded',
                       url='', release_date='', artist='', album=''):
    if not db_path:
        return

    with sqlite3.connect(db_path) as conn:
        if add_id:
            try:
                # Inject artist and album dynamically into the database
                conn.execute(
                    """
                    INSERT INTO downloads (id, media_type, quality, file_format, quality_met, bit_depth, 
                    sampling_rate, saved_path, url, release_date, status, artist, album) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
                    (item_id, media_type, quality, file_format, quality_met, bit_depth, sampling_rate,
                     saved_path, url, release_date, status, artist, album),
                )
                conn.commit()
            except sqlite3.IntegrityError:
                # Provide clean visual feedback instead of an error
                logger.info(f"{YELLOW}[i] Already in database, skipping.{OFF}")
            except sqlite3.Error as e:
                logger.error(f"{RED}Unexpected DB error: {e}{OFF}")
        else:
            return conn.execute(
                "SELECT id FROM downloads WHERE id=? AND quality=?",
                (item_id, quality),
            ).fetchone()
 
 
def get_stats(db_path):
    """Returns a list of unique artists from the database."""
    if not db_path:
        return []
    try:
        with sqlite3.connect(db_path) as conn:
            cursor = conn.cursor()
            # We select unique artists, excluding empty strings
            cursor.execute("SELECT DISTINCT artist FROM downloads WHERE artist != '' ORDER BY artist ASC")
            return [row[0] for row in cursor.fetchall()]
    except sqlite3.Error:
        return []            


import os
import logging
import time
from mutagen.flac import FLAC
from mutagen.id3 import ID3
from qobuz_dl.db import handle_download_id
from qobuz_dl.color import GREEN, RED, YELLOW, CYAN, OFF

logger = logging.getLogger(__name__)

def sync_database(directory, db_path, client):
    """
    Scans the local directory and restores missing Qobuz IDs into the local DB.
    Uses embedded custom tags or falls back to Reverse Lookup via Qobuz API with anti-ban delay.
    """
    logger.info(f"\n{YELLOW}[*] Starting Local Database Synchronization...{OFF}")
    logger.info(f"{YELLOW}[*] Scanning directory: {directory}{OFF}")

    # --- PATCH OS.WALK: Immune a parentesi quadre e case-insensitive ---
    all_files = []
    for root, _, files in os.walk(directory):
        for file in files:
            if file.lower().endswith(('.flac', '.mp3')):
                all_files.append(os.path.join(root, file))
    # -------------------------------------------------------------------

    if not all_files:
        logger.info(f"{YELLOW}[!] No audio files found in {directory}.{OFF}")
        return

    logger.info(f"{YELLOW}[*] Found {len(all_files)} audio files. Processing tags...{OFF}")

    added_tracks = 0
    added_albums = set()

    try:
        for file_path in all_files:
            track_id = None
            album_id = None
            isrc = None
            quality = 27
            file_format = "FLAC" if file_path.lower().endswith(".flac") else "MP3"

            try:
                if file_path.lower().endswith(".flac"):
                    audio = FLAC(file_path)
                    track_id = audio.get("QOBUZTRACKID", [None])[0]
                    album_id = audio.get("QOBUZALBUMID", [None])[0]
                    isrc = audio.get("isrc", [None])[0]
                elif file_path.lower().endswith(".mp3"):
                    audio = ID3(file_path)
                    track_txxx = audio.get("TXXX:QOBUZTRACKID")
                    if track_txxx: track_id = track_txxx.text[0]
                    album_txxx = audio.get("TXXX:QOBUZALBUMID")
                    if album_txxx: album_id = album_txxx.text[0]
                    tsrc = audio.get("TSRC")
                    if tsrc: isrc = tsrc.text[0]
                
                # --- REVERSE LOOKUP VIA API FOR OLD FILES ---
                if not track_id and isrc:
                    logger.info(f"{CYAN}[*] Missing local ID. Fetching via API (ISRC: {isrc})...{OFF}")
                    res = client.search_tracks(isrc, limit=1)
                    if res and "tracks" in res and res["tracks"]["items"]:
                        q_track = res["tracks"]["items"][0]
                        track_id = str(q_track["id"])
                        album_id = str(q_track.get("album", {}).get("id", ""))
                    
                    # Human behavior delay to prevent Qobuz API throttling and hanging
                    time.sleep(0.2)
                
                # Inject Track ID into DB
                if track_id:
                    handle_download_id(
                        db_path=db_path, item_id=track_id, add_id=True, media_type="track",
                        quality=quality, file_format=file_format, saved_path=file_path
                    )
                    added_tracks += 1
                
                # Inject Album ID into DB
                if album_id and album_id not in added_albums:
                    handle_download_id(
                        db_path=db_path, item_id=album_id, add_id=True, media_type="album",
                        quality=quality, file_format=file_format, saved_path=os.path.dirname(file_path)
                    )
                    added_albums.add(album_id)

            except Exception as e:
                logger.error(f"{RED}[!] Error processing {file_path}: {e}{OFF}")

    except KeyboardInterrupt:
        logger.warning(f"\n{YELLOW}[!] Synchronization forcibly interrupted by user!{OFF}")
        logger.warning(f"{YELLOW}[!] Don't worry, all progress up to this point has been safely saved.{OFF}")

    logger.info(f"{GREEN}[+] Sync complete! Restored {added_tracks} tracks and {len(added_albums)} albums into the local database.{OFF}")


from .lyrics_engine import LyricsEngine
import logging
import os
import sys
import time
import random
import subprocess
import re
import threading
import signal
from typing import Tuple
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

import requests
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from pathvalidate import sanitize_filename, sanitize_filepath
from tqdm import tqdm

import qobuz_dl.metadata as metadata
from qobuz_dl.color import OFF, GREEN, RED, YELLOW, CYAN
from qobuz_dl.exceptions import NonStreamable
from qobuz_dl.settings import QobuzDLSettings
from qobuz_dl.utils import get_album_artist, clean_filename
from qobuz_dl.db import handle_download_id
from qobuz_dl.constants import DEFAULT_FOLDER, DEFAULT_TRACK, DEFAULT_MULTIPLE_DISC_TRACK, OK_MAX_CHARACTER_LENGTH

# UI Lock to prevent text scrambling during multithreading
print_lock = threading.Lock()

# Global Abort Event for graceful CTRL+C handling and file unlock
abort_event = threading.Event()

def safe_print(*args, **kwargs):
    with print_lock:
        text = " ".join(map(str, args))
        end = kwargs.get('end', '\n')
        tqdm.write(text, end=end)

# --- FIX ISSUE #216: Normalize Release Type ---
def format_release_type(release_type: str) -> str:
    """
    Normalizes the release type from Qobuz APIs.
    Converts 'ep' to 'EP', 'single' to 'Single', 'album' to 'Album', etc.
    Returns 'Unknown' as a robust fallback if data is missing.
    """
    if not release_type:
        return "Unknown"
    
    release_type = release_type.lower()
    if release_type == "ep":
        return "EP"
        
    return release_type.title()
# --------------------------------------------------------

def process_folder_format_with_subdirs(folder_format, attr_dict, path=None, legacy_charmap=False):
    path_parts = folder_format.split('/')
    cleaned_parts = []
    for part in path_parts:
        if part:
            try:
                formatted_part = part.format(**attr_dict)
                # AGGIUNTO legacy_charmap QUI:
                cleaned_part = sanitize_filepath(clean_filename(formatted_part, legacy_charmap=legacy_charmap), replacement_text="_")
                
                # --- FIX: SMART TRUNCATION FOR ALBUM FOLDER ---
                if cleaned_part and len(cleaned_part) > 120:
                    start_f = cleaned_part[:60].rstrip(' ."-_\'')
                    end_f = cleaned_part[-50:].lstrip(' ."-_\'')
                    cleaned_part = f"{start_f}...{end_f}"
                    
                if cleaned_part:
                    cleaned_parts.append(cleaned_part)
            except KeyError as e:
                logger.warning(f"{YELLOW}Format error ({e}), using original text.{OFF}")
                # AGGIUNTO legacy_charmap ANCHE QUI:
                cleaned_part = sanitize_filepath(clean_filename(part, legacy_charmap=legacy_charmap), replacement_text="_")
                
                if cleaned_part and len(cleaned_part) > 120:
                    start_f = cleaned_part[:60].rstrip(' ."-_\'')
                    end_f = cleaned_part[-50:].lstrip(' ."-_\'')
                    cleaned_part = f"{start_f}...{end_f}"
                    
                if cleaned_part:
                    cleaned_parts.append(cleaned_part)
    
    final_path = os.path.join(*cleaned_parts) if cleaned_parts else ""
    if path is not None:
        return os.path.join(path, final_path)
    return final_path

QL_DOWNGRADE = "FormatRestrictedByFormatAvailability"
DEFAULT_FORMATS = {
    "MP3": [
        "{album_artist} - {album_title} ({year}) [MP3]",
        "{track_number} - {track_title}",
    ],
    "Unknown": [
        "{album_artist} - {album_title}",
        "{track_number} - {track_title}",
    ],
}

EMB_COVER_NAME = "embed_cover.jpg"

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Download:
    def __init__(
        self,
        client,
        item_id: str,
        path: str,
        quality: int,
        embed_art: bool = False,
        albums_only: bool = False,
        downgrade_quality: bool = False,
        cover_og_quality: bool = False,
        no_cover: bool = False,
        folder_format=None,
        track_format=None,
        fetch_lyrics: bool = False,
        no_lrc_files: bool = False,
        genius_token: str = None,
        no_credits: bool = False,
        settings: QobuzDLSettings = None,
        download_db=None,
        is_playlist: bool = False,           
        playlist_track_number: int = None, 
        booklet_only: bool = False,        
    ):
        self.client = client
        self.item_id = item_id
        self.path = path
        self.quality = quality
        self.albums_only = albums_only
        self.embed_art = embed_art
        self.downgrade_quality = downgrade_quality
        self.cover_og_quality = cover_og_quality
        self.no_cover = no_cover
        self.folder_format = folder_format or DEFAULT_FOLDER
        self.track_format = track_format or DEFAULT_TRACK
        self.no_credits = no_credits
        self.booklet_only = booklet_only        

        self.fetch_lyrics = fetch_lyrics
        self.no_lrc_files = no_lrc_files
        if self.fetch_lyrics:
            self.lyrics_engine = LyricsEngine(genius_token)

        self.settings = settings or QobuzDLSettings()
        self.download_db = download_db
        
        self.is_playlist = is_playlist                       
        self.playlist_track_number = playlist_track_number   
        
        self._original_folder_format = folder_format or DEFAULT_FOLDER
        self._original_track_format = track_format or DEFAULT_TRACK
        self._original_multiple_disc_track_format = settings.multiple_disc_track_format if settings else DEFAULT_MULTIPLE_DISC_TRACK

    def download_id_by_type(self, track=True):
        self.folder_format = self._original_folder_format
        self.track_format = self._original_track_format
        if self.settings:
            self.settings.multiple_disc_track_format = self._original_multiple_disc_track_format
        
        if not track:
            self.download_release()
        else:
            self.download_track()

    def download_release(self):
        count = 0
        album_meta = self.client.get_album_meta(self.item_id)

        if not album_meta.get("streamable"):
            raise NonStreamable("This release is not streamable")

        if self.albums_only and (
            album_meta.get("release_type") != "album"
            or album_meta.get("artist").get("name") == "Various Artists"
        ):
            logger.info(f'{OFF}Ignoring Single/EP/VA: {album_meta.get("title", "n/a")}')
            return

        album_title = _get_title(album_meta)
        url = album_meta.get("url", "")
        release_date = album_meta.get("release_date_original", "")

        format_info = self._get_format(album_meta)
        file_format, quality_met, bit_depth, sampling_rate = format_info

        if not self.downgrade_quality and not quality_met:
            logger.info(f"{OFF}Skipping {album_title} as it doesn't meet quality requirement")
            return

        logger.info(
            f"\n{YELLOW}Downloading: {album_title}\nQuality: {file_format}"
            f" ({bit_depth}/{sampling_rate})\n{OFF}"
        )
        
        album_attr = self._get_album_attr(
            album_meta, album_title, file_format, bit_depth, sampling_rate
        )
        
        self._determine_formats(album_meta=album_meta, album_attr=album_attr, tracks_meta=album_meta["tracks"]["items"],
                                track_attr=None, is_track=False, file_format=file_format, settings=self.settings)
        
        legacy_flag = getattr(self.settings, 'legacy_charmap', False) if hasattr(self, 'settings') else False
        target_dirn = process_folder_format_with_subdirs(self.folder_format, album_attr, self.path, legacy_charmap=legacy_flag)
        base_path, folder_name = os.path.split(target_dirn)
        
        incomplete_dirn = os.path.join(base_path, f"[INCOMPLETE] {folder_name}")
        inprogress_dirn = os.path.join(base_path, f"[IN PROGRESS] {folder_name}")
        
        is_standard_album = not getattr(self, 'is_playlist', False)
        
        if is_standard_album:
            working_dirn = inprogress_dirn
            try:
                if os.path.exists(incomplete_dirn):
                    os.rename(incomplete_dirn, working_dirn)
                elif os.path.exists(target_dirn):
                    os.rename(target_dirn, working_dirn)
            except OSError as e:
                logger.warning(f"{YELLOW}[!] Could not rename existing folder to [IN PROGRESS]. Operating in standard mode. ({e}){OFF}")
                working_dirn = target_dirn
        else:
            working_dirn = target_dirn
            
        os.makedirs(working_dirn, exist_ok=True)
        dirn = working_dirn

        media_count = album_meta.get("media_count", 1)
        is_multiple = True if media_count > 1 else False
        
        delay_time = getattr(self.settings, 'delay', 0)
        if delay_time == 0 and '--delay' in sys.argv:
            try: delay_time = int(sys.argv[sys.argv.index('--delay') + 1])
            except: pass
            
        active_workers = int(getattr(self.settings, 'max_workers', 3))
        is_parallel = False
        
        if delay_time > 0:
            safe_print(f"{YELLOW}[*] Safety Delay active: Multithreading disabled (Sequential mode).{OFF}")
            active_workers = 1
        elif active_workers > 1:
            is_parallel = True
            safe_print(f"{YELLOW}[*] Multithreading Enabled ({active_workers} workers): UI optimized for clean parallel logging.{OFF}")
            
        failed_tracks = 0
        aborted_by_user = False
        abort_event.clear()

        # --- SIGINT HIJACKER (Hacker Fix) ---
        # Intercept Ctrl+C to prevent core/cli from brutally killing the process,
        # ensuring we have time to release file locks and rename the folder to [INCOMPLETE].
        original_sigint = None
        try:
            original_sigint = signal.getsignal(signal.SIGINT)
            def custom_sigint_handler(sig, frame):
                abort_event.set()
                raise KeyboardInterrupt
            signal.signal(signal.SIGINT, custom_sigint_handler)
        except Exception:
            pass

        try:
            self._generate_tracklist(album_meta, dirn, album_title, file_format, bit_depth, sampling_rate)

            if self.settings.no_cover:
                logger.info(f"{OFF}Skipping cover")
            else:
                _get_extra(album_meta["image"]["large"], dirn, art_size=self.settings.saved_art_size)

            if self.settings.embed_art:
                _get_extra(album_meta["image"]["large"], dirn, extra=EMB_COVER_NAME, art_size=self.settings.embedded_art_size)

            if "goodies" in album_meta:
                _download_goodies(album_meta, dirn)
                
            if getattr(self, 'booklet_only', False):
                safe_print(f"{YELLOW}[*] --booklet-only flag active. Skipping audio tracks.{OFF}")
                
                if is_standard_album and working_dirn == inprogress_dirn:
                    try:
                        os.rename(working_dirn, incomplete_dirn)
                    except OSError as e:
                        logger.warning(f"{YELLOW}[!] Impossibile rinominare la cartella in [INCOMPLETE]. ({e}){OFF}")
                
                return
             
            with concurrent.futures.ThreadPoolExecutor(max_workers=active_workers) as executor:
                futures = []
                for i in album_meta["tracks"]["items"]:
                    if abort_event.is_set():
                        break
                    try:
                        parse = self.client.get_track_url(i["id"], fmt_id=self.quality)
                    except Exception as e:
                        safe_print(f"{RED}[!] API Error for track {i.get('track_number', 'unknown')} (ID: {i['id']}): {e}{OFF}")
                        safe_print(f"{YELLOW}[*] Skipping track and continuing with the album...{OFF}")
                        count += 1
                        failed_tracks += 1
                        continue

                    if "sample" not in parse and parse["sampling_rate"]:
                        is_mp3 = True if int(self.quality) == 5 else False
                        futures.append(
                            executor.submit(
                                self._download_and_tag, dirn, count, parse, i, album_meta, False,
                                is_mp3, i.get("media_number") if is_multiple else None, is_parallel=is_parallel
                            )
                        )
                    else:
                        logger.info(f"{OFF}Demo. Skipping")
                        failed_tracks += 1
                    count += 1
                    
                try:
                    for f in futures:
                        while not f.done():
                            if abort_event.is_set():
                                break
                            time.sleep(0.2)
                            
                    if not abort_event.is_set():
                        for f in futures:
                            try:
                                res = f.result()
                                if res is False:
                                    failed_tracks += 1
                            except Exception as inner_e:
                                safe_print(f"{RED}[!] Track download failed: {inner_e}{OFF}")
                                failed_tracks += 1
                except (KeyboardInterrupt, SystemExit):
                    abort_event.set()
                    aborted_by_user = True
                    safe_print(f"\n{RED}[!] CTRL+C Intercepted: Securing files and folders...{OFF}")
                    
            if not aborted_by_user:
                _clean_embed_art(dirn, self.settings)
                if getattr(self, 'fetch_lyrics', False) and not self.no_credits:
                    self._append_lyrics_to_booklet(dirn, album_title)
                    
        except (KeyboardInterrupt, SystemExit):
            abort_event.set()
            aborted_by_user = True
            safe_print(f"\n{RED}[!] CTRL+C Intercepted: Securing files and folders...{OFF}")
            
        finally:
            # Restore original signal handler so the CLI functions normally afterwards
            try:
                if original_sigint:
                    signal.signal(signal.SIGINT, original_sigint)
            except Exception:
                pass
                
        if aborted_by_user:
            # Crucial: Wait for threads to drop OS file locks before attempting folder rename
            time.sleep(1.5)
            
        # --- FINAL DIRECTORY STATE EVALUATION ---
        if is_standard_album and working_dirn == inprogress_dirn:
            final_dirn = target_dirn if (failed_tracks == 0 and not aborted_by_user) else incomplete_dirn
            try:
                os.rename(working_dirn, final_dirn)
            except OSError as e:
                logger.warning(f"{YELLOW}[!] Could not rename final folder state (OS Lock might still be active). ({e}){OFF}")
                final_dirn = working_dirn
            
            if aborted_by_user:
                safe_print(f"{YELLOW}[!] Download aborted. Folder successfully marked as [INCOMPLETE].{OFF}")
            elif failed_tracks > 0:
                safe_print(f"\n{YELLOW}[!] Album downloaded partially ({failed_tracks} tracks skipped). Folder marked as [INCOMPLETE].{OFF}")
        else:
            final_dirn = working_dirn
        
        if aborted_by_user:
            os._exit(1)
            
        # --- DATABASE UPGRADE: Inject artist and album metadata ---
        db_artist = album_attr.get("album_artist", "Unknown")
        db_album = album_attr.get("album_title", "Unknown")
        
        handle_download_id(db_path=self.download_db, item_id=self.item_id, add_id=True, media_type="album",
                           quality=self.quality, file_format=file_format, quality_met=quality_met,
                           bit_depth=bit_depth, sampling_rate=sampling_rate, saved_path=final_dirn,
                           url=url, release_date=release_date, artist=db_artist, album=db_album)
        safe_print(f"{GREEN}Completed{OFF}")

    def download_track(self):
        parse = self.client.get_track_url(self.item_id, self.quality)
        if "sample" not in parse and parse["sampling_rate"]:
            track_meta = self.client.get_track_meta(self.item_id)
            
            if getattr(self, 'is_playlist', False) and getattr(self, 'playlist_track_number', None):
                track_meta["track_number"] = self.playlist_track_number
            
            track_title = _get_title(track_meta)
            artist = _safe_get(track_meta, "performer", "name")
            logger.info(f"\n{YELLOW}Downloading: {artist} - {track_title}{OFF}")
            url = track_meta.get("album", {}).get("url", "")
            release_date = track_meta.get("release_date_original", "")
            format_info = self._get_format(track_meta, is_track_id=True, track_url_dict=parse)
            file_format, quality_met, bit_depth, sampling_rate = format_info

            folder_format, track_format = _clean_format_str(self.folder_format, self.track_format, str(bit_depth))

            if not self.downgrade_quality and not quality_met:
                logger.info(f"{OFF}Skipping {track_title} as it doesn't meet quality requirement")
                return
                
            track_attr = self._get_track_attr(
                track_meta, track_title, bit_depth, sampling_rate, file_format
            )
            
            self._determine_formats(album_meta=track_meta.get("album", {}), album_attr=None, tracks_meta=[track_meta],
                                    track_attr=track_attr, is_track=True, file_format=file_format, settings=self.settings)
            
            legacy_flag = getattr(self.settings, 'legacy_charmap', False) if hasattr(self, 'settings') else False
            dirn = process_folder_format_with_subdirs(self.folder_format, track_attr, self.path, legacy_charmap=legacy_flag)
            os.makedirs(dirn, exist_ok=True)

            if getattr(self, 'is_playlist', False):
                logger.info(f"{OFF}Skipping standard cover save to keep playlist folder clean")
            elif self.settings.no_cover:
                logger.info(f"{OFF}Skipping cover")
            else:
                _get_extra(track_meta["album"]["image"]["large"], dirn, art_size=self.settings.saved_art_size)

            if self.settings.embed_art:
                embed_path = os.path.join(dirn, EMB_COVER_NAME)
                if os.path.exists(embed_path):
                    try:
                        os.remove(embed_path)
                    except OSError:
                        pass
                
                _get_extra(track_meta["album"]["image"]["large"], dirn, extra=EMB_COVER_NAME,
                           art_size=self.settings.embedded_art_size)
            else:
                logger.info(f"{OFF}Skipping embedded art")
                
            is_mp3 = True if int(self.quality) == 5 else False
            
            self._download_and_tag(
                dirn,
                1,
                parse,
                track_meta,
                track_meta,
                True,
                is_mp3,
                False,
                is_parallel=False
            )
            
            _clean_embed_art(dirn, self.settings)
            
            # --- DATABASE UPGRADE: Inject artist and album metadata ---
            db_artist = track_attr.get("artist", "Unknown")
            db_album = track_attr.get("album", "Unknown")
            
            handle_download_id(db_path=self.download_db, item_id=self.item_id, add_id=True, media_type="track",
                               quality=self.quality, file_format=file_format, quality_met=quality_met,
                               bit_depth=bit_depth, sampling_rate=sampling_rate, saved_path=dirn,
                               url=url, release_date=release_date, artist=db_artist, album=db_album)
        else:
            logger.info(f"{OFF}Demo. Skipping")
        logger.info(f"{GREEN}Completed{OFF}")

    def _download_and_tag(
        self,
        root_dir,
        tmp_count,
        track_url_dict,
        track_metadata,
        album_or_track_metadata,
        is_track,
        is_mp3,
        multiple=None,
        is_parallel=False
    ) -> bool:
        extension = ".mp3" if is_mp3 else ".flac"

        track_artist = _safe_get(track_metadata, "performer", "name")
        filename_attr = self._get_filename_attr(
            track_artist,
            track_metadata,
            album_or_track_metadata.get("album", {}) if is_track else album_or_track_metadata
        )

        # --- FIX MARROBHD & SYNC-PLAYLIST: CLEAN PLAYLIST NAMING ---
        legacy_flag = getattr(self.settings, 'legacy_charmap', False) if hasattr(self, 'settings') else False
        
        if getattr(self, 'is_playlist', False):
            # Forza un nome file pulito senza numero traccia per le playlist
            clean_playlist_format = "{artist} - {track_title}"
            formatted_path = sanitize_filename(clean_filename(clean_playlist_format.format(**filename_attr), legacy_charmap=legacy_flag), replacement_text="_")
        elif multiple and self.settings.multiple_disc_one_dir:
            formatted_path = sanitize_filename(clean_filename(self.settings.multiple_disc_track_format.format(**filename_attr), legacy_charmap=legacy_flag), replacement_text="_")
        else:
            # FIX MULTI-DISC PATHING: Includiamo la cartella CD nel percorso finale se ci sono più dischi
            base_formatted = sanitize_filename(clean_filename(self.track_format.format(**filename_attr), legacy_charmap=legacy_flag), replacement_text="_")
            total_discs = album_or_track_metadata.get('media_count', 1)
            if multiple and total_discs > 1:
                try: d_num = int(multiple) if not isinstance(multiple, bool) else 1
                except (ValueError, TypeError): d_num = 1
                disc_folder = f"{self.settings.multiple_disc_prefix} {d_num:02}"
                formatted_path = os.path.join(disc_folder, base_formatted)
            else:
                formatted_path = base_formatted
        # -----------------------------------------------------------
            
        max_len = 180
        if len(formatted_path) > max_len:
            start_part = formatted_path[:110].rstrip(' ."-_\'')
            end_part = formatted_path[-60:].lstrip(' ."-_\'')
            formatted_path = f"{start_part}...{end_part}"
            
        final_file = os.path.join(root_dir, formatted_path) + extension

        if os.path.exists(final_file):
            safe_print(f"{CYAN}[*] Skipping: {os.path.basename(final_file)} (Already exists){OFF}")
            return True

        if abort_event.is_set():
            return False

        time.sleep(1)
        try:
            url = track_url_dict["url"]
        except KeyError:
            logger.info(f"{OFF}Track not available for download")
            return False

        total_discs = album_or_track_metadata.get('media_count', 1)
        if multiple and total_discs > 1 and (not self.settings.multiple_disc_one_dir):
            try:
                d_num = int(multiple) if not isinstance(multiple, bool) else 1
            except (ValueError, TypeError): 
                d_num = 1
            root_dir = os.path.join(root_dir, f"{self.settings.multiple_disc_prefix} {d_num:02}")
        
        if not os.path.exists(root_dir):
            os.makedirs(root_dir, exist_ok=True)

        filename = os.path.join(root_dir, f"~tmp_{tmp_count:02}.tmp")
        track_title = track_metadata.get("title")
        track_no = str(track_metadata.get('track_number', 0)).zfill(2)
        desc = f"{track_no}. {track_title}"

        FALLBACK_TIERS = [27, 7, 6, 5]
        TIER_NAMES = {27: "24-bit/>96kHz", 7: "24-bit/96kHz", 6: "16-bit/44.1kHz (CD)", 5: "MP3 320kbps"}
        
        try:
            start_idx = FALLBACK_TIERS.index(int(self.quality))
        except ValueError:
            start_idx = 0
            
        qualities_to_try = FALLBACK_TIERS[start_idx:]
        success = False
        final_fmt = int(self.quality)

        for attempt_fmt in qualities_to_try:
            if abort_event.is_set():
                return False
                
            if attempt_fmt != int(self.quality):
                safe_print(f"{YELLOW}[!] Automatic downgrade: Attempting to save in {TIER_NAMES[attempt_fmt]}...{OFF}")

            def get_fresh_url(fmt=attempt_fmt, force_segments=False):
                return self.client.get_track_url(track_metadata["id"], fmt_id=fmt, force_segments=force_segments)

            try:
                fresh_track_dict = get_fresh_url(force_segments=False)
                
                if "url" in fresh_track_dict:
                    try:
                        tqdm_download(fresh_track_dict["url"], filename, desc, is_parallel=is_parallel)
                        success = True
                        final_fmt = attempt_fmt
                        break
                    except Exception as e:
                        if abort_event.is_set(): return False
                        safe_print(f"{YELLOW}[!] Akamai block detected. Activating fallback segmented download...{OFF}")
                        fresh_track_dict = get_fresh_url(force_segments=True)
                
                if "url_template" in fresh_track_dict:
                    tqdm_download_segments(fresh_track_dict, filename, desc, is_parallel=is_parallel)
                    success = True
                    final_fmt = attempt_fmt
                    break
                elif not success:
                    raise Exception("No valid format returned by the server.")

            except Exception as e:
                pass

        if not success and not abort_event.is_set():
            safe_print(f"\n{RED}[!] TRACK {track_no} DEFINITIVELY DISCARDED AFTER ALL DOWNGRADES.{OFF}")
            safe_print(f"{YELLOW}[!] Skipping to the next track...{OFF}\n")
            return False
            
        if abort_event.is_set():
            return False

        is_mp3 = True if final_fmt == 5 else False
        extension = ".mp3" if is_mp3 else ".flac"

        tag_function = metadata.tag_mp3 if is_mp3 else metadata.tag_flac
        try:
            tag_function(
                filename, root_dir, final_file, track_metadata,
                album_or_track_metadata, is_track, self.embed_art, settings=self.settings,
            )
        except Exception as e:
            safe_print(f"{RED}[!] Error tagging: {e}{OFF}")

        if getattr(self, 'fetch_lyrics', False) and hasattr(self, 'lyrics_engine') and not abort_event.is_set():
            album_artist = _safe_get(track_metadata, "album", "artist", "name")
            performer_name = _safe_get(track_metadata, "performer", "name") or _safe_get(track_metadata, "artist", "name", default="Unknown")
            search_artist = performer_name if album_artist in [None, "Various Artists"] else album_artist

            search_album = _safe_get(track_metadata, "album", "title", default="")
            
            with print_lock:                           
                self.lyrics_engine.fetch_and_inject(
                    file_path=final_file, 
                    artist=search_artist, 
                    track=track_title, 
                    album=search_album,
                    save_lrc=not self.no_lrc_files,
                    embed_lyrics=getattr(self.settings, 'embed_lyrics', True)
                )

        delay_time = getattr(self.settings, 'delay', 0)
        if delay_time == 0 and '--delay' in sys.argv:
            try: delay_time = int(sys.argv[sys.argv.index('--delay') + 1])
            except: pass
            
        if delay_time > 0 and not abort_event.is_set():
            safe_print(f"{YELLOW}[*] Sleeping for {delay_time} seconds to prevent rate limiting...{OFF}")
            time.sleep(delay_time)
            
        return True

    @staticmethod
    def _get_filename_attr(track_artist, track_metadata: dict, album_metadata: dict):     
        def _flatten_artists(artist_data):
            if isinstance(artist_data, list): return ", ".join(artist_data)
            return str(artist_data) if artist_data else ""
           
        album_artist_raw = get_album_artist(album_metadata)
        album_artist_str = _flatten_artists(album_artist_raw) if album_artist_raw else track_artist

        return {
            "artist": track_artist,
            "albumartist": album_artist_str,
            "tracktitle": _get_title(track_metadata),
            "album_title":  _get_title(album_metadata),
            "album_title_base": album_metadata.get("title"),
            "album_artist": album_artist_str,
            "track_id": track_metadata.get("id"),
            "track_artist": track_artist,
            "track_composer": _safe_get(track_metadata,"composer", "name"),
            "track_number": f'{track_metadata.get("track_number", 0):02}',
            "isrc": track_metadata.get("isrc"),
            "bit_depth": track_metadata.get("maximum_bit_depth"),
            "sampling_rate": track_metadata.get("maximum_sampling_rate"),
            "track_title": _get_title(track_metadata),
            "track_title_base": track_metadata.get("title"),
            "version": track_metadata.get("version"),
            "year": track_metadata.get("release_date_original", "").split("-")[0],
            "disc_number": f'{track_metadata.get("media_number"):02}',
            "release_date": track_metadata.get("release_date_original"),
            "ExplicitFlag": "[E]" if track_metadata.get("parental_warning") else "",
            "explicit": "[E]" if track_metadata.get("parental_warning") else "",
        }

    @staticmethod
    def _get_track_attr(meta, track_title, bit_depth, sampling_rate, file_format):
        album_meta = meta.get("album", {})
        def _flatten_artists(artist_data):
            if isinstance(artist_data, list): return ", ".join(artist_data)
            return str(artist_data) if artist_data else ""
            
        album_artist_raw = get_album_artist(album_meta)
        album_artist_str = _flatten_artists(album_artist_raw) if album_artist_raw else _safe_get(meta, "performer", "name")

        return {
            "album": _get_title(album_meta),
            "artist": album_artist_str,
            "tracktitle": track_title,
            "track_title": track_title,
            "track_title_base": meta.get("title", ""),
            "album_id": meta.get("id", ""),
            "album_url": meta.get("url", ""),
            "album_title": _get_title(album_meta),
            "album_title_base": album_meta.get("title", ""),
            "album_artist": album_artist_str,
            "album_genre": meta.get("genre", {}).get("name", ""),
            "album_composer": meta.get("composer", {}).get("name", ""),
            "label": re.sub(r'\s*[\;\/]\s*|\s+\-\s+',' ／ ', ' '.join(meta.get("label",{}).get("name", "").split())).strip(),
            "copyright": meta.get("copyright", ""),
            "upc": meta.get("upc", ""),
            "barcode": meta.get("upc", ""),
            "release_date": meta.get("release_date_original", ""),
            "year": meta.get("release_date_original", "").split("-")[0],
            "media_type": meta.get("product_type", "").capitalize(),
            "format": file_format,
            "bit_depth": bit_depth,
            "sampling_rate": sampling_rate,
            "album_version": meta.get("version", ""),
            "version_tag": f" - {meta.get('version')}" if meta.get("version") else "",
            "disc_count": meta.get("media_count", ""),
            "track_count": meta.get("track_count", ""),
            "ExplicitFlag": "[E]" if album_meta.get("parental_warning") else "",
            "explicit": "[E]" if album_meta.get("parental_warning") else "",
            "release_type": format_release_type(album_meta.get("release_type")),
        }

    @staticmethod
    def _get_album_attr(meta, album_title, file_format, bit_depth, sampling_rate):
        def _flatten_artists(artist_data):
            if isinstance(artist_data, list): return ", ".join(artist_data)
            return str(artist_data) if artist_data else ""
            
        album_artist_raw = get_album_artist(meta)
        album_artist_str = _flatten_artists(album_artist_raw)

        return {
            "artist": meta.get("artist", {}).get("name", ""),
            "album": album_title,
            "album_id": meta.get("id", ""),
            "album_url": meta.get("url", ""),
            "album_title": album_title,
            "album_title_base": meta.get("title", ""),
            "album_artist": album_artist_str,
            "album_genre": meta.get("genre", {}).get("name", ""),
            "album_composer": meta.get("composer", {}).get("name", ""),
            "label": re.sub(r'\s*[\;\/]\s*|\s+\-\s+',' ∕ ', ' '.join(meta.get("label",{}).get("name", "").split())).strip(),
            "copyright": meta.get("copyright", ""),
            "upc": meta.get("upc", ""),
            "barcode": meta.get("upc", ""),
            "release_date": meta.get("release_date_original", ""),
            "year": meta.get("release_date_original", "").split("-")[0],
            "media_type": meta.get("product_type", "").capitalize(),
            "format": file_format,
            "bit_depth": bit_depth,
            "sampling_rate": sampling_rate,
            "album_version": meta.get("version", ""),
            "version_tag": f" - {meta.get('version')}" if meta.get("version") else "",
            "disc_count": meta.get("media_count", 1),
            "track_count": meta.get("track_count", 1),
            "ExplicitFlag": "[E]" if meta.get("parental_warning") else "",
            "explicit": "[E]" if meta.get("parental_warning") else "",
            "release_type": format_release_type(meta.get("release_type")),
        }

    def _get_format(self, item_dict, is_track_id=False, track_url_dict=None):
        # Aggiungi questa protezione anti-crash SOLO per le release complete (Album/EP)
        if not is_track_id:
            if "tracks" not in item_dict or not item_dict["tracks"].get("items"):
                from qobuz_dl.exceptions import NonStreamable
                raise NonStreamable("This release has no tracks available (possibly region-locked or removed)")

        # FIX: Applica la logica corretta a seconda se è una traccia singola o un album
        track_dict = item_dict if is_track_id else item_dict["tracks"]["items"][0]
        
        # INIZIALIZZAZIONE MANCANTE: Di default la qualità è rispettata!
        quality_met = True
        
        try:
                     
            # FIX PRINCIPALE: Inizializziamo sempre una variabile interna
            new_track_dict = (
                self.client.get_track_url(track_dict["id"], fmt_id=self.quality)
                if not track_url_dict
                else track_url_dict
            )
            
            # Se la chiamata API non ha restituito un dizionario valido (es. None), forziamo l'eccezione
            if not new_track_dict:
                 raise KeyError("No URL dict returned by API")

            restrictions = new_track_dict.get("restrictions")
            if isinstance(restrictions, list):
                if any(restriction.get("code") == QL_DOWNGRADE for restriction in restrictions):
                    quality_met = False

            return ("FLAC", quality_met, new_track_dict["bit_depth"], new_track_dict["sampling_rate"])
            
        except (KeyError, requests.exceptions.HTTPError, Exception):
            # In caso di errore (geoblocco, traccia non disponibile, ecc.), restituiamo i valori None
            # in modo che il downloader "salti" la traccia senza mandare in crash l'intero loop.
            return ("Unknown", quality_met, None, None)

    def _determine_formats(self, album_meta, album_attr, tracks_meta, track_attr, is_track, file_format, settings: QobuzDLSettings):
        format_combinations = [
            (self._original_folder_format, self._original_track_format, self._original_multiple_disc_track_format),
            (settings.fallback_folder_format, self._original_track_format, self._original_multiple_disc_track_format),
            (settings.fallback_folder_format, DEFAULT_TRACK, DEFAULT_MULTIPLE_DISC_TRACK),
            (DEFAULT_FOLDER, DEFAULT_TRACK, DEFAULT_MULTIPLE_DISC_TRACK)
        ]

        media_count = album_meta.get("media_count", 1)
        is_multiple = True if media_count > 1 else False
        extension = ".flac" if file_format.lower() == "flac" else ".mp3"
        
        # --- NEW: Retrieve legacy charmap flag ---
        legacy_flag = getattr(settings, 'legacy_charmap', False) if settings else False

        for folder_fmt, track_fmt, multi_disc_fmt in format_combinations:
            folder_fmt, track_fmt = _clean_format_str(folder_fmt, track_fmt, file_format)
            valid_combination = True
            
            try:
                if is_track:
                    root_dir = process_folder_format_with_subdirs(folder_fmt, track_attr, legacy_charmap=legacy_flag)
                else:
                    root_dir = process_folder_format_with_subdirs(folder_fmt, album_attr, legacy_charmap=legacy_flag)

                for track_metadata in tracks_meta:
                    track_artist = _safe_get(track_metadata, "performer", "name")
                    filename_attr = self._get_filename_attr(track_artist, track_metadata, album_meta)

                    curr_root_dir = root_dir
                    if is_multiple and self.settings.multiple_disc_one_dir:
                        track_path = sanitize_filename(clean_filename(multi_disc_fmt.format(**filename_attr), legacy_charmap=legacy_flag), replacement_text="_")
                    else:
                        if is_multiple and not self.settings.multiple_disc_one_dir:
                            disc_dir = f"{self.settings.multiple_disc_prefix} {track_metadata['media_number']:02}"
                            curr_root_dir = os.path.join(root_dir, disc_dir)
                        
                        track_path = sanitize_filename(clean_filename(track_fmt.format(**filename_attr), legacy_charmap=legacy_flag), replacement_text="_")

                    # --- FIX: REMOVED OBSOLETE LENGTH CHECK ---
                    # We no longer invalidate the user's custom format if the path is too long.
                    # String truncation is now safely handled downstream in _download_and_tag.

            except (KeyError, ValueError):
                # Fallback to the next format ONLY if there is a missing tag/variable in the metadata
                valid_combination = False
                continue

            if valid_combination:
                self.folder_format = folder_fmt
                self.track_format = track_fmt
                if self.settings:
                    self.settings.multiple_disc_track_format = multi_disc_fmt
                return

        self.folder_format = DEFAULT_FOLDER
        self.track_format = DEFAULT_TRACK

    def _generate_tracklist(self, meta, dirn, album_title, file_format, bit_depth, sampling_rate):
        import re
        import textwrap
        
        if self.no_credits or abort_event.is_set():
            return
        
        safe_title = sanitize_filename(album_title)
        tracklist_path = os.path.join(dirn, f"{safe_title} - Tracklist.txt")
        
        if os.path.isfile(tracklist_path):
            return
            
        safe_print(f"{CYAN}[+] Generating Digital Booklet...{OFF}")
        
        artist_name = _safe_get(meta, "artist", "name", default="Unknown Artist")
        composer = _safe_get(meta, "composer", "name", default="N/A")
        label = _safe_get(meta, "label", "name", default="Independent")
        raw_genre = _safe_get(meta, "genre", "name", default="Unknown Genre")
        genre = metadata.LOCAL_GENRE_MAP.get(raw_genre, raw_genre) if raw_genre != "Unknown Genre" else raw_genre
        release_date = meta.get("release_date_original", "Unknown Date")
        
        try:
            with open(tracklist_path, "w", encoding="utf-8") as f:
                f.write("=" * 70 + "\n")
                f.write(f"ALBUM      : {album_title}\n")
                if composer != "N/A": f.write(f"COMPOSER   : {composer}\n")
                f.write(f"MAIN ART.  : {artist_name}\n")
                f.write(f"LABEL      : {label}\n")
                f.write(f"GENRE      : {genre}\n")
                f.write(f"RELEASE    : {release_date}\n")
                f.write(f"QUALITY    : {file_format} ({bit_depth}-Bit / {sampling_rate} kHz)\n")
                f.write("=" * 70 + "\n\n")
                
                tracks = meta.get("tracks", {}).get("items", [])
                total_discs = max((track.get("media_number", 1) for track in tracks), default=1)
                current_disc = None 
                
                for track in tracks:
                    disc_num = track.get("media_number", 1)
                    if total_discs > 1 and disc_num != current_disc:
                        if current_disc is not None: f.write("\n")
                        f.write(f"--- DISC {disc_num} ---\n\n")
                        current_disc = disc_num

                    t_num = str(track.get("track_number", 0)).zfill(2)
                    t_title = track.get("title", "Unknown Title")
                    duration = int(track.get("duration", 0))
                    mins, secs = divmod(duration, 60)
                    dur_str = f"[{mins:02}:{secs:02}]"
                    
                    f.write(f"{f'{t_num}. {t_title}':<60} {dur_str}\n")
                    
                    performers_raw = track.get("performers", "")
                    if performers_raw:
                        for line in re.split(r'\r?\n|\s+-\s+', str(performers_raw)):
                            if line.strip(): f.write(f"    * {line.strip()}\n")
                    else:
                        t_artist = _safe_get(track, "performer", "name", default=artist_name)
                        f.write(f"    {t_artist}\n")
                    f.write("\n")
                
                description = meta.get("description")
                if description:
                    f.write("\n" + "=" * 70 + "\nALBUM REVIEW / NOTES\n" + "=" * 70 + "\n\n")
                    clean_desc = re.sub(r'<[^<]+>', '', re.sub(r'<br\s*/?>', '\n', str(description)))
                    for p in clean_desc.split('\n'):
                        if p.strip():
                            f.write(textwrap.fill(p.strip(), width=70) + "\n\n")

            safe_print(f"{GREEN}  L Completed: Digital Booklet.txt (Credits & Review){OFF}")
        except Exception as e:
            safe_print(f"{RED}[!] Error creating booklet: {e}{OFF}")

    def _append_lyrics_to_booklet(self, dirn, album_title):
        import re
        if abort_event.is_set(): return
        
        safe_title = sanitize_filename(album_title)
        tracklist_path = os.path.join(dirn, f"{safe_title} - Tracklist.txt")
        if not os.path.isfile(tracklist_path): return
            
        audio_files = []
        for root, _, files in os.walk(dirn):
            for f in files:
                if f.lower().endswith(('.flac', '.mp3')):
                    audio_files.append(os.path.join(root, f))
                    
        audio_files.sort()
        lyrics_to_append = []
        
        for audio_path in audio_files:
            base_path = os.path.splitext(audio_path)[0]
            lrc_path, txt_path = f"{base_path}.lrc", f"{base_path}.txt" 
            base_name = os.path.basename(base_path)
            lyrics_text = ""
            
            if os.path.exists(lrc_path):
                with open(lrc_path, "r", encoding="utf-8") as f:
                    raw_lyrics = f.read()
                clean_lyrics = re.sub(r'\[[a-zA-Z]+:.*?\]\n?|\[\d{2,}:\d{2}\.\d{2,3}\]', '', raw_lyrics)
                clean_lines = [line.strip() for line in clean_lyrics.splitlines() if line.strip() or (lyrics_text and lyrics_text[-1] != "")]
                lyrics_text = "\n".join(clean_lines).strip()
            elif os.path.exists(txt_path) and "Tracklist" not in txt_path:
                with open(txt_path, "r", encoding="utf-8") as f:
                    lyrics_text = f.read().strip()
                    
            if lyrics_text:
                lyrics_to_append.append(f"--- {base_name} ---\n\n{lyrics_text}\n\n")
                
        if lyrics_to_append:
            try:
                with open(tracklist_path, "a", encoding="utf-8") as f:
                    f.write("\n" + "=" * 70 + "\nALBUM LYRICS\n" + "=" * 70 + "\n\n")
                    f.writelines(lyrics_to_append)
                safe_print(f"{CYAN}[+] Lyrics cleanly formatted and appended to Digital Booklet.{OFF}")
            except Exception as e:
                logger.error(f"{RED}[!] Error appending lyrics to booklet: {e}{OFF}")

def _get_description(item: dict, track_title, multiple=None):
    downloading_title = f"{track_title} [{item.get('bit_depth', '')}/{item.get('sampling_rate', '')}]"
    if multiple:
        downloading_title = f"[CD {multiple}] {downloading_title}"
    return downloading_title

def tqdm_download(url_or_callable, fname, track_name, is_parallel=False):
    if abort_event.is_set(): return
    G, Y, C, O = "\033[92m", "\033[93m", "\033[96m", "\033[0m"

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36",
        "Accept": "audio/webm,audio/ogg,audio/wav,audio/*;q=0.9,*/*;q=0.5",
        "Connection": "keep-alive"
    }

    if not is_parallel:
        safe_print(f"{C}[+] In progress: {track_name}{O}")
        tqdm_desc = f" {G}Downloading{O}"
        b_format = "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"
    else:
        tqdm_desc = ""
        b_format = ""

    downloaded_size = 0
    total_size = 0
    max_retries = 5
    backoff_delays = [2, 4, 8, 16, 32] 

    for attempt in range(max_retries):
        if abort_event.is_set(): return
        try:
            url = url_or_callable() if callable(url_or_callable) else url_or_callable

            if downloaded_size > 0:
                headers['Range'] = f'bytes={downloaded_size}-'
                mode = 'ab'
            else:
                headers['Range'] = 'bytes=0-'
                mode = 'wb'
            
            with requests.Session() as s:
                r = s.get(url, allow_redirects=True, stream=True, headers=headers, timeout=(10, 60))
                
                if r.status_code == 416: return 
                if r.status_code not in [200, 206]:
                    raise Exception(f"Status Server: {r.status_code}")

                if total_size == 0:
                    total_size = downloaded_size + int(r.headers.get('content-length', 0))

                if is_parallel and downloaded_size == 0 and attempt == 0:
                    size_mb = total_size / (1024 * 1024)
                    safe_print(f"{C}[+] In progress: {track_name} [{size_mb:.1f} MB]{O}")

                with open(fname, mode) as file, tqdm(
                    total=total_size, unit="iB", unit_scale=True, unit_divisor=1024,
                    desc=tqdm_desc, initial=downloaded_size, bar_format=b_format, leave=False, disable=is_parallel
                ) as bar:
                    for data in r.iter_content(chunk_size=65536):
                        if abort_event.is_set(): return
                        if data:
                            size = file.write(data)
                            downloaded_size += size
                            if not is_parallel:
                                bar.update(size)
            
            if downloaded_size >= total_size:
                safe_print(f"{G}  L Completed: {track_name}{O}")
                return 

        except Exception as e:
            if "404" in str(e):
                if os.path.exists(fname): os.remove(fname)
                raise Exception("HTTP 404: File not found on server.")
                        
            if attempt < max_retries - 1:
                wait = backoff_delays[attempt]
                safe_print(f"\n{Y}[!] Server block. Retrying in {wait}s ({attempt+1}/{max_retries}) | Error details: {e}{O}")
                time.sleep(wait)
            else:
                if os.path.exists(fname): os.remove(fname)
                raise Exception(f"Definitive timeout after {max_retries} attempts. Last error: {e}")

    if downloaded_size < total_size and not abort_event.is_set():
        if os.path.exists(fname): os.remove(fname)
        raise Exception("Incomplete download")

def _get_title(item_dict):
    item_title = item_dict.get("title")
    version = item_dict.get("version")
    if version:
        item_title = f"{item_title} ({version})" if version.lower() not in item_title.lower() else item_title
    return item_title


def _get_extra(item, dirn, extra="cover.jpg", art_size=None, og_quality=False):
    if abort_event.is_set(): return
    extra_file = os.path.join(dirn, extra)
    if os.path.isfile(extra_file):
        logger.info(f"{OFF}{extra} was already downloaded")
        return
        
    if og_quality: art_size = "org"
    if art_size in ["50", "100", "150", "300", "600", "max", "org"]:
        item = item.replace("_600.", f"_{art_size}.")
        
    try:
        tqdm_download(item, extra_file, extra, is_parallel=False)
    except Exception as e:
        safe_print(f"  {YELLOW}[!] Skipping cover art '{extra}': URL unreachable ({e}){OFF}")

def _clean_format_str(folder: str, track: str, file_format: str) -> Tuple[str, str]:
    final = []
    for i, fs in enumerate((folder, track)):
        if fs.endswith(".mp3"): fs = fs[:-4]
        elif fs.endswith(".flac"): fs = fs[:-5]
        fs = fs.strip()

        if file_format in ("MP3", "Unknown") and ("bit_depth" in fs or "sampling_rate" in fs):
            default = DEFAULT_FORMATS[file_format][i]
            logger.error(f"{RED}invalid format string for format {file_format}. defaulting to {default}")
            fs = default
        final.append(fs)
    return tuple(final)

def _safe_get(d: dict, *keys, default=None):
    curr = d
    res = default
    for key in keys:
        res = curr.get(key, default)
        if res == default or not hasattr(res, "__getitem__"):
            return res
        else:
            curr = res
    return res

def tqdm_download_segments(track_url_dict, fname, track_name, is_parallel=False):
    if abort_event.is_set(): return
    G, C, O = "\033[92m", "\033[96m", "\033[0m" 
    
    tmp_fname = fname + ".mp4"
    n_segments = track_url_dict["n_segments"]
    url_template = track_url_dict["url_template"]
    raw_key = track_url_dict["raw_key"]

    def get_seg_size(seg_num):
        if abort_event.is_set(): return 0
        url = url_template.replace("$SEGMENT$", str(seg_num))
        try:
            r = requests.head(url, timeout=5)
            return int(r.headers.get("content-length", 0))
        except: return 0

    total_size = 0
    with ThreadPoolExecutor(max_workers=8) as ex:
        futures_size = [ex.submit(get_seg_size, i) for i in range(n_segments + 1)]
        for f in futures_size:
            while not f.done():
                if abort_event.is_set(): return
                time.sleep(0.1)
            total_size += f.result()

    if is_parallel:
        size_mb = total_size / (1024 * 1024)
        safe_print(f"{C}[+] In progress: {track_name} [{size_mb:.1f} MB]{O}")
        tqdm_desc, b_format = "", ""
    else:
        safe_print(f"{C}[+] In progress: {track_name}{O}")
        tqdm_desc = f" {G}Segmented Download{O}"
        b_format = "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"

    def fetch_segment_fluid(seg_num):
        if abort_event.is_set(): return bytearray()
        url = url_template.replace("$SEGMENT$", str(seg_num))
        r = requests.get(url, stream=True, timeout=15)
        r.raise_for_status()
        seg_data = bytearray()
        
        for chunk in r.iter_content(chunk_size=65536):
            if abort_event.is_set(): return bytearray()
            seg_data.extend(chunk)
            if not is_parallel:
                bar.update(len(chunk)) 
        return seg_data

    try:
        with open(tmp_fname, "wb") as file, tqdm(
            total=total_size, unit="iB", unit_scale=True, unit_divisor=1024,
            desc=tqdm_desc, bar_format=b_format, leave=False, disable=is_parallel
        ) as bar:

            segment_uuid = None
            for i in range(2):
                seg_data = fetch_segment_fluid(i)
                if abort_event.is_set(): return
                if i == 1:
                    segment_uuid = _get_qobuz_segment_uuid(seg_data)
                    if segment_uuid is None:
                        raise ConnectionError(f"Cannot find segment UUID for {fname}")

                file.write(_decrypt_qobuz_segment(seg_data, raw_key, segment_uuid))

            if n_segments >= 2:
                with ThreadPoolExecutor(max_workers=8) as executor:
                    futures_seg = [executor.submit(fetch_segment_fluid, i) for i in range(2, n_segments + 1)]
                    for f in futures_seg:
                        while not f.done():
                            if abort_event.is_set(): return
                            time.sleep(0.2)
                        seg_data = f.result()
                        if not abort_event.is_set():
                            file.write(_decrypt_qobuz_segment(seg_data, raw_key, segment_uuid))

        if abort_event.is_set(): return
        if not is_parallel:
            safe_print(f" {G}  > Assembling the final FLAC file...{O}")
            
        remux = subprocess.run(["ffmpeg", "-nostdin", "-v", "error", "-y", "-i", tmp_fname, "-c:a", "copy", "-f", "flac", fname], stdout=subprocess.DEVNULL, stderr=subprocess.PIPE, text=True)
        if remux.returncode != 0:
            raise ConnectionError(f"FFmpeg remux failed for {fname}")
        
        safe_print(f"{G}  L Completed: {track_name}{O}")

    finally:
        if os.path.isfile(tmp_fname):
            try: os.remove(tmp_fname)
            except OSError: pass


def _get_qobuz_segment_uuid(segment_data):
    pos = 0
    while pos + 24 <= len(segment_data):
        size = int.from_bytes(segment_data[pos : pos + 4], "big")
        if size <= 0 or pos + size > len(segment_data): break

        if bytes(segment_data[pos + 4 : pos + 8]) == b"uuid":
            return bytes(segment_data[pos + 8 : pos + 24])
        pos += size
    return None


def _decrypt_qobuz_segment(segment_data, raw_key, segment_uuid):
    if segment_uuid is None: return bytes(segment_data)

    buf = bytearray(segment_data)
    pos = 0
    while pos + 8 <= len(buf):
        size = int.from_bytes(buf[pos : pos + 4], "big")
        if size <= 0 or pos + size > len(buf): break

        if bytes(buf[pos + 4 : pos + 8]) == b"uuid" and bytes(buf[pos + 8 : pos + 24]) == segment_uuid:
            pointer = pos + 28
            data_end = pos + int.from_bytes(buf[pointer : pointer + 4], "big")
            pointer += 4
            counter_len = buf[pointer]
            pointer += 1
            frame_count = int.from_bytes(buf[pointer : pointer + 3], "big")
            pointer += 3

            for _ in range(frame_count):
                frame_len = int.from_bytes(buf[pointer : pointer + 4], "big")
                pointer += 6
                flags = int.from_bytes(buf[pointer : pointer + 2], "big")
                pointer += 2
                frame_start, data_end = data_end, data_end + frame_len

                if flags:
                    counter = bytes(buf[pointer : pointer + counter_len]) + (b"\x00" * (16 - counter_len))
                    decryptor = Cipher(algorithms.AES(raw_key), modes.CTR(counter)).decryptor()
                    buf[frame_start:data_end] = decryptor.update(bytes(buf[frame_start:data_end])) + decryptor.finalize()
                pointer += counter_len
        pos += size
    return bytes(buf)

def _download_goodies(album_meta, dirn):
    if abort_event.is_set(): return
    try:
        for goody in album_meta.get("goodies", []):
            if abort_event.is_set(): break
            if not goody.get("url"): continue
            goody_name = sanitize_filename(clean_filename(f'{album_meta.get("title")} ({goody.get("id")}).pdf'))
            _get_extra(goody.get("url"), dirn, extra=goody_name)
    except Exception as e:
        logger.error(f"{RED}Error downloading goodies: {e}", exc_info=True)


def _clean_embed_art(dirn, settings=None):
    embed_file = os.path.join(dirn, EMB_COVER_NAME)
    if os.path.exists(embed_file):
        try:
            time.sleep(0.5) 
            os.remove(embed_file)
        except OSError:
            pass


import os
import configparser
import urllib.request
import xml.etree.ElementTree as ET
import questionary
from qobuz_dl.qopy import Client
from qobuz_dl.color import GREEN, YELLOW, RED, CYAN, OFF

def setup_client(config_path, config, section):
    """Initializes the Qobuz client reading from the config file."""
    app_id = config.get(section, 'app_id')
    secrets = config.get(section, 'secrets')
    auth_token = config.get(section, 'auth_token')
    email = config.get(section, 'email', fallback="") or None
    pwd = config.get(section, 'password', fallback="") or None

    api = Client(email, pwd, user_auth_token=auth_token, app_id=app_id, secrets=secrets)
    api.auth_token = auth_token
    return api

def get_or_save_rss_link(config_path, config, section):
    """Retrieves the RSS link or asks the user for it on first run."""
    if config.has_option(section, 'musicbutler_rss'):
        rss_link = config.get(section, 'musicbutler_rss').strip()
        if rss_link:
            return rss_link

    print(f"{YELLOW}[!] No RSS feed found.{OFF}")
    rss_link = questionary.text("Paste your private MusicButler RSS link here:").ask()
    
    if rss_link:
        config.set(section, 'musicbutler_rss', rss_link)
        with open(config_path, 'w') as configfile:
            config.write(configfile)
        print(f"{GREEN}[+] Link permanently saved to config!{OFF}\n")
    
    return rss_link

def fetch_rss_releases(rss_url):
    """Downloads and parses the RSS/Atom feed ignoring XML namespaces."""
    print(f"{CYAN}[*] Syncing with MusicButler...{OFF}")
    try:
        req = urllib.request.Request(rss_url, headers={'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)'})
        with urllib.request.urlopen(req) as response:
            xml_data = response.read()
            
        root = ET.fromstring(xml_data)
        releases = []
        
        for elem in root.iter():
            tag = elem.tag.split('}')[-1] 
            
            if tag in ['item', 'entry']:
                for child in elem.iter():
                    child_tag = child.tag.split('}')[-1]
                    if child_tag == 'title' and child.text:
                        releases.append(child.text.strip())
                        break
                        
        return releases
    except Exception as e:
        print(f"{RED}[!] Error reading RSS feed: {e}{OFF}")
        return []

def run_radar():
    """Main execution function for the radar command."""
    appdata_path = os.getenv('APPDATA')
    config_path = os.path.join(appdata_path, 'qobuz-dl', 'config.ini')
    
    config = configparser.ConfigParser()
    config.read(config_path)
    if not config.sections():
        print(f"{RED}[!] config.ini file not found at {config_path}{OFF}")
        return
        
    section = config.sections()[0]
    
    # 1. RSS Link Management
    rss_url = get_or_save_rss_link(config_path, config, section)
    if not rss_url:
        print(f"{RED}[!] Operation cancelled. No link provided.{OFF}")
        return

    # 2. Connect to Qobuz API
    try:
        api = setup_client(config_path, config, section)
    except Exception as e:
        print(f"{RED}[!] Connection error to Qobuz: {e}{OFF}")
        return

    # 3. Download and parse RSS
    releases = fetch_rss_releases(rss_url)
    
    if not releases:
        print(f"{YELLOW}[!] No new releases found in the feed.{OFF}")
        return
        
    print(f"{GREEN}[+] Found {len(releases)} new releases! Searching on Qobuz...{OFF}\n")

    # 4. Search on Qobuz and prepare the UI menu
    choices = []
    for release_title in releases:
        search_result = api.search_albums(release_title, limit=1)
        
        if search_result and "albums" in search_result and search_result["albums"]["items"]:
            album_data = search_result["albums"]["items"][0]
            album_id = album_data["id"]
            
            artist = album_data.get("artist", {}).get("name", "Unknown")
            title = album_data.get("title", "Unknown")
            display_name = f"{artist} - {title}"
            
            choices.append(questionary.Choice(title=display_name, value=album_id))
        else:
            print(f"{YELLOW}[!] Not found on Qobuz: {release_title}{OFF}")

    if not choices:
        print(f"{YELLOW}\n[!] None of the releases in the feed are currently available on Qobuz.{OFF}")
        return

    # 5. Interactive UI Menu
    print("\n")
    selected_album_ids = questionary.checkbox(
        "🎧 Select releases to add to Favorites (Space to select, Enter to confirm):",
        choices=choices
    ).ask()

    if not selected_album_ids:
        print(f"{YELLOW}[*] No albums selected. Exiting.{OFF}")
        return

    # 6. Add to Favorites
    print(f"\n{CYAN}[*] Adding {len(selected_album_ids)} albums to favorites...{OFF}")
    for album_id in selected_album_ids:
        try:
            api.add_favorite_album(album_id)
            print(f"{GREEN}  [+] Added: ID {album_id}{OFF}")
        except Exception as e:
            print(f"{RED}  [-] Error with ID {album_id}: {e}{OFF}")
            
    print(f"\n{GREEN}✅ Operation complete! You can now run qobuz-dl to download them.{OFF}")


# Default format strings
DEFAULT_FOLDER = "{album_artist} - {album_title} ({year}) [{format} {bit_depth}]"
DEFAULT_TRACK = "{track_number} - {track_title_base}"
DEFAULT_MULTIPLE_DISC_TRACK = "{disc_number}.{track_number} - {track_title_base}"
# character length for the longest allowed album track filename (folder_name + track_name + extension).
OK_MAX_CHARACTER_LENGTH = 180


import base64
import hashlib
import logging
import time
import unicodedata
import json

import requests
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from qobuz_dl.exceptions import (
    AuthenticationError,
    InvalidAppIdError,
    InvalidAppSecretError,
    InvalidQuality,
)
from qobuz_dl.color import GREEN, YELLOW, RED, OFF, RESET

try:
    from qobuz_dl.bundle import Bundle
except ImportError:
    Bundle = None

logger = logging.getLogger(__name__)

class Client:
    def __init__(self, email, pwd, app_id, secrets, user_auth_token=None, force_english=True):
        logger.info(f"{YELLOW}Logging...{OFF}")
        self.secrets = secrets
        self.id = str(app_id)
        self.secrets = secrets
        self.force_english = force_english
        
        if Bundle:
            try:
                b = Bundle()
                fresh_id = str(b.get_app_id())
                if fresh_id:
                    self.id = fresh_id
                    self.secrets = list(b.get_secrets().values())
                    logger.info(f"{GREEN}[+] App ID dynamically updated: {self.id}{OFF}")
            except Exception:
                pass

        self.session = requests.Session()
        
        # --- CONDITIONAL ENGLISH LANGUAGE OVERRIDE ---
        if self.force_english:
            self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36",
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en-US,en;q=0.9",
            "X-App-Language": "en",
            "X-App-Region": "US",
            "Sec-Ch-Ua": "\"Chromium\";v=\"124\", \"Google Chrome\";v=\"124\", \"Not-A.Brand\";v=\"99\"",
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Ch-Ua-Platform": "\"Windows\"",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-site",
            "X-App-Id": self.id,
        })
        # ---------------------------------------------
        
        self.session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36",
            "X-App-Id": self.id,
        })
        self.base = "https://www.qobuz.com/api.json/0.2/"
        self.sec = None
# Variables for encryption session management
        self.session_id = None
        self.session_infos = None
        self.session_key = None
        
        self.uat = None
        self.force_english = force_english 
        
        self.auth(email, pwd, user_auth_token)
        self.cfg_setup()

    def api_call(self, epoint, **kwargs):
        if epoint == "user/login":
            if "user_auth_token" in kwargs and kwargs["user_auth_token"]:
                params = {
                    "user_auth_token": kwargs["user_auth_token"],
                    "app_id": self.id,
                }
                logger.info(f"{YELLOW}Trying to login with user_auth_token{OFF}")
            else:
                params = {
                    "email": kwargs["email"],
                    "password": kwargs["pwd"],
                    "app_id": self.id,
                }
                logger.info(f"{YELLOW}Trying to login with email/password{OFF}")
            
            # add debug info
            logger.info(f"{YELLOW}Login params: {params}{OFF}")
        elif epoint == "track/get":
            params = {"track_id": kwargs["id"]}
        elif epoint == "album/get":
            params = {"album_id": kwargs["id"]}
        elif epoint == "playlist/get":
            params = {
                "extra": "tracks",
                "playlist_id": kwargs["id"],
                "limit": 500,
                "offset": kwargs["offset"],
            }
        elif epoint == "artist/get":
            params = {
                "app_id": self.id,
                "artist_id": kwargs["id"],
                "limit": 500,
                "offset": kwargs["offset"],
                "extra": "albums",
            }
        elif epoint == "label/get":
            params = {
                "label_id": kwargs["id"],
                "limit": 500,
                "offset": kwargs["offset"],
                "extra": "albums",
            }
        elif epoint == "favorite/getUserFavorites":
            unix = int(time.time())
            r_sig = "favoritegetUserFavorites" + str(unix) + kwargs.get("sec", self.sec)
            r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
            params = {
                "app_id": self.id,
                "user_auth_token": getattr(self, 'uat', None),
                "type": kwargs.get("fav_type", "albums"), 
                "limit": kwargs.get("limit", 100),
                "offset": kwargs.get("offset", 0),
                "request_ts": unix,
                "request_sig": r_sig_hashed,
            }
        elif epoint == "track/getFileUrl":
            unix = time.time()
            track_id = kwargs["id"]
            fmt_id = kwargs["fmt_id"]
            if int(fmt_id) not in (5, 6, 7, 27):
                raise InvalidQuality("Invalid quality id: choose between 5, 6, 7 or 27")
            r_sig = "trackgetFileUrlformat_id{}intentstreamtrack_id{}{}{}".format(
                fmt_id, track_id, unix, kwargs.get("sec", self.sec)
            )
            r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
            params = {
                "request_ts": unix,
                "request_sig": r_sig_hashed,
                "track_id": track_id,
                "format_id": fmt_id,
                "intent": "stream",
            }
        else:
            params = kwargs
            
        r = self.session.get(self.base + epoint, params=params)
        
        if epoint == "user/login":
            if r.status_code == 401:
                raise AuthenticationError("Invalid credentials.\n" + RESET)
            elif r.status_code == 400:
                raise InvalidAppIdError("Invalid app id.\n" + RESET)
            else:
                logger.info(f"{GREEN}Logged: OK{OFF}")
        elif (
            epoint in ["track/getFileUrl", "favorite/getUserFavorites"]
            and r.status_code == 400
        ):
            raise InvalidAppSecretError(f"Invalid app secret: {r.json()}.\n" + RESET)
        r.raise_for_status()
        
        # Unicode Normalization for JSON strings
        json_data = r.json()
        return self._normalize_json_strings(json_data)

    def _normalize_json_strings(self, obj):
        """Recursively normalize Unicode strings in JSON objects (NFC form)"""
        if isinstance(obj, str):
            # --- WINDOWS PATH FIX: Convert '...' to Unicode Ellipsis (U+2026) ---
            # Avoid modifying URL links (which contain '://')
            if "..." in obj and "://" not in obj:
                obj = obj.replace("...", "…")
            # --------------------------------------------------------------------
            return unicodedata.normalize('NFC', obj)
        elif isinstance(obj, dict):
            return {k: self._normalize_json_strings(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._normalize_json_strings(item) for item in obj]
        else:
            return obj

    def auth(self, email, pwd, user_auth_token=None):
        # If the token is present, skip the password!
        if user_auth_token:
            self.uat = user_auth_token
        elif len(pwd) > 60:
            self.uat = pwd
        else:
            usr_info = self.api_call("user/login", email=email, pwd=pwd)
            if not usr_info.get("user", {}).get("credential", {}).get("parameters"):
                logger.info(f"{YELLOW}[!] Free account detected or validation bypassed.{OFF}")
            self.uat = usr_info["user_auth_token"]
        
        self.session.headers.update({"X-User-Auth-Token": self.uat})
        
        try:
            user_info = self.api_call("user/get")
            cred = user_info.get("credential") or user_info.get("user", {}).get("credential", {})
            self.label = cred.get("parameters", {}).get("short_label", "Studio")
            
            # --- FIX: Save user ID strictly required for favorites ---
            self.user_id = user_info.get("id") or user_info.get("user", {}).get("id")
            # -------------------------------------------------------------------------
            
            logger.info(f"{GREEN}Logged: OK (Membership: {self.label}){OFF}")
        except Exception:
            logger.info(f"{YELLOW}[!] Profile validation bypassed.{OFF}")
            self.label = "Studio"
            self.user_id = None

    # NEW CRYPTOGRAPHIC FUNCTIONS (Patch 0004)
    def _modern_sig(self, epoint, params, sec):
        object_, method = epoint.split("/")
        r_sig = [object_, method]
        for key in sorted(params):
            value = params[key]
            if key not in ("request_ts", "request_sig") and isinstance(
                value, (str, int, float)
            ):
                r_sig.extend((key, str(value)))
        r_sig.extend((str(params["request_ts"]), sec))
        return hashlib.md5("".join(r_sig).encode("utf-8")).hexdigest()

    @staticmethod
    def _b64url_decode(value):
        return base64.urlsafe_b64decode(value + "=" * (-len(value) % 4))

    def _derive_session_key(self):
        salt, info = self.session_infos.split(".")
        return HKDF(
            algorithm=hashes.SHA256(),
            length=16,
            salt=self._b64url_decode(salt),
            info=self._b64url_decode(info),
        ).derive(bytes.fromhex(self.sec))

    def _unwrap_track_key(self, key_token):
        _, wrapped, iv = key_token.split(".")
        decryptor = Cipher(
            algorithms.AES(self.session_key),
            modes.CBC(self._b64url_decode(iv)),
        ).decryptor()
        padded = decryptor.update(self._b64url_decode(wrapped)) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        return unpadder.update(padded) + unpadder.finalize()

    # NEW API_CALL ENGINE
    def api_call(self, epoint, **kwargs):
        if epoint == "user/login":
            if "user_auth_token" in kwargs and kwargs["user_auth_token"]:
                params = {
                    "user_auth_token": kwargs["user_auth_token"],
                    "app_id": self.id,
                }
            else:
                params = {
                    "email": kwargs["email"],
                    "password": kwargs["pwd"],
                    "app_id": self.id,
                }
        elif epoint == "track/getFileUrl":
            track_id = kwargs["id"]
            fmt_id = kwargs["fmt_id"]
            if int(fmt_id) not in (5, 6, 7, 27):
                raise InvalidQuality("Invalid quality id: choose between 5, 6, 7 or 27")
            params = {
                "track_id": track_id,
                "format_id": fmt_id,
                "intent": "stream",
            }
            # Use the old string method for MP3 compatibility
            unix = int(time.time())
            sec_to_use = kwargs.get('sec', self.sec)
            r_sig = f"trackgetFileUrlformat_id{fmt_id}intentstreamtrack_id{track_id}{unix}{sec_to_use}"
            params["request_ts"] = unix
            params["request_sig"] = hashlib.md5(r_sig.encode()).hexdigest()

        elif epoint == "session/start":
            params = {"profile": "qbz-1"}
            params["request_ts"] = int(time.time())
            params["request_sig"] = self._modern_sig(
                epoint, params, kwargs.get("sec", self.sec)
            )
        elif epoint == "file/url":
            track_id = kwargs["id"]
            fmt_id = kwargs["fmt_id"]
            if int(fmt_id) not in (6, 7, 27):
                raise InvalidQuality("Invalid quality id: choose between 6, 7 or 27")
            params = {
                "track_id": track_id,
                "format_id": fmt_id,
                "intent": "import",
            }
            params["request_ts"] = int(time.time())
            params["request_sig"] = self._modern_sig(
                epoint, params, kwargs.get("sec", self.sec)
            )
        elif epoint == "favorite/getUserFavorites":
            unix = int(time.time())
            r_sig = "favoritegetUserFavorites" + str(unix) + kwargs.get("sec", self.sec)
            r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
            params = {
                "app_id": self.id,
                "user_auth_token": getattr(self, 'uat', None),
                "user_id": getattr(self, 'user_id', None), 
                "type": kwargs.get("fav_type", "albums"),
                "limit": kwargs.get("limit", 100),
                "offset": kwargs.get("offset", 0),
                "request_ts": unix,
                "request_sig": r_sig_hashed,
            }
        else:
            # Restore behavior for standard calls like album/get
            params = {'app_id': self.id}
            
            # --- CONDITIONAL ENGLISH PARAMS OVERRIDE ---
            if getattr(self, 'force_english', True):
                params['lang'] = 'en'
                params['locale'] = 'en_US'
            # -------------------------------------------
            
            val_id = kwargs.get('id')
            for k, v in kwargs.items():
                if k not in ['id', 'sec', 'fmt_id']:
                    params[k] = v

            if epoint == "album/get": params["album_id"] = val_id
            elif epoint == "track/get": params["track_id"] = val_id
            elif epoint == "playlist/get": params["playlist_id"] = val_id; params["extra"] = "tracks"
            elif epoint == "artist/get": params["artist_id"] = val_id; params["extra"] = "albums"
            elif epoint == "label/get": params["label_id"] = val_id; params["extra"] = "albums"

        # PATCH: Added favorite/create to POST methods
        if epoint in ["user/login", "favorite/create"]:
            r = self.session.post(self.base + epoint, data=params)
        elif epoint == "session/start":
            r = self.session.post(
                self.base + epoint,
                data=params,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
        else:
            r = self.session.get(self.base + epoint, params=params)

        if epoint == "user/login" and r.status_code == 400:
            if "invalid" in r.text.lower():
                raise AuthenticationError("Invalid email or password.")
            else:
                logger.info(f"{GREEN}Logged: OK{OFF}")
        elif (
            epoint in ["track/getFileUrl", "favorite/getUserFavorites", "file/url"]
            and r.status_code == 400
        ):
            raise InvalidAppSecretError(f"Invalid app secret: {r.json()}.\n" + RESET)
        
        if epoint == "user/get" and r.status_code == 400: return {}
        r.raise_for_status()
        
        # Apply string normalizer to the network call output
        return self._normalize_json_strings(r.json())

    def multi_meta(self, epoint, key, id, type):
        offset = 0
        limit = 50
        
        while True:
            j = self.api_call(epoint, id=id, offset=offset, limit=limit, type=type)
            res = j[type] if type and type in j else j
            
            items_key = "tracks" if "playlist" in epoint else "albums"
            items = res.get(items_key, {}).get("items", [])
            
            if not items:
                break
                
            yield res
            
            offset += len(items)
            total_available = res.get(items_key, {}).get("total", res.get(key, 0))
            if offset >= total_available:
                break

    # --- METADATA FUNCTIONS (Do not delete!) ---
    def get_track_meta(self, id): 
        return self.api_call("track/get", id=id)

    # --- NEW LAST.FM FUNCTIONS ---
    def get_track_ids_from_list(self, tracks_list: list) -> list:
        from qobuz_dl.color import OFF, GREEN, RED, YELLOW, CYAN
        import difflib
        
        print(f"{CYAN}[*] Matching Last.fm tracks with Qobuz database (Fuzzy matching & Interactive mode enabled)...{OFF}")
        valid_track_ids = []
        
        AUTO_ACCEPT_THRESHOLD = 0.75 
        PROMPT_THRESHOLD = 0.60      
        
        for item in tracks_list:
            target_artist = item['artist'].lower()
            target_title = item['title'].lower()
            query = f"{item['artist']} {item['title']}"
            
            try:
                search_results = self.search_tracks(query, limit=5)
                
                best_match_id = None
                best_match_name = ""
                highest_ratio = 0.0
                
                if search_results and "tracks" in search_results and search_results["tracks"]["items"]:
                    for q_track in search_results["tracks"]["items"]:
                        q_artist_raw = q_track.get("performer", {}).get("name", "Unknown")
                        q_title_raw = q_track.get("title", "Unknown")
                        
                        q_artist = q_artist_raw.lower()
                        q_title = q_title_raw.lower()
                        
                        target_str = f"{target_artist} {target_title}"
                        q_str = f"{q_artist} {q_title}"
                        
                        ratio = difflib.SequenceMatcher(None, target_str, q_str).ratio()
                        
                        if ratio > highest_ratio:
                            highest_ratio = ratio
                            best_match_id = q_track["id"]
                            best_match_name = f"{q_artist_raw} - {q_title_raw}"
                    
                    if highest_ratio >= AUTO_ACCEPT_THRESHOLD and best_match_id:
                        valid_track_ids.append(best_match_id)
                        
                    elif highest_ratio >= PROMPT_THRESHOLD and best_match_id:
                        print(f"\n{YELLOW}[?] Borderline match detected ({highest_ratio*100:.0f}% similarity){OFF}")
                        print(f"    Target (Last.fm): {item['artist']} - {item['title']}")
                        print(f"    Found  (Qobuz)  : {best_match_name}")
                        
                        choice = input(f"{CYAN}    Do you want to download this track anyway? [y/n]: {OFF}").strip().lower()
                        
                        if choice == 'y':
                            valid_track_ids.append(best_match_id)
                            print(f"{GREEN}    [+] Track accepted manually.{OFF}")
                        else:
                            print(f"{RED}    [-] Track skipped manually.{OFF}")
                            
                    else:
                        print(f"{YELLOW}[!] Skipping: '{query}' (Best match was only {highest_ratio*100:.0f}% similar){OFF}")
                        
                else:
                    print(f"{YELLOW}[!] Skipping (No results on Qobuz for): '{query}'{OFF}")
                    
            except Exception as e:
                print(f"{RED}[!] Error searching for '{query}': {e}{OFF}")
                
        print(f"\n{GREEN}[+] Successfully matched {len(valid_track_ids)} out of {len(tracks_list)} tracks!{OFF}")
        return valid_track_ids

    # --- SEARCH FUNCTIONS (Crash-Proof) ---
    def search_albums(self, query, limit=20):
        try: return self.api_call("catalog/search", query=query, type="albums", limit=limit)
        except Exception: return {}

    def search_tracks(self, query, limit=20):
        try: return self.api_call("catalog/search", query=query, type="tracks", limit=limit)
        except Exception: return {}

    def search_playlists(self, query, limit=20):
        try: return self.api_call("catalog/search", query=query, type="playlists", limit=limit)
        except Exception: return {}

    def search_artists(self, query, limit=20):
        try: return self.api_call("catalog/search", query=query, type="artists", limit=limit)
        except Exception: return {}

    # --- NEW FAVORITES FUNCTION ---
    def get_favorites(self, fav_type="albums", limit=100, offset=0):
        """
        Fetches user favorites dynamically. 
        fav_type can be: 'albums', 'tracks', 'artists', 'playlists'
        """
        try: 
            return self.api_call("favorite/getUserFavorites", fav_type=fav_type, limit=limit, offset=offset)
        except Exception as e: 
            logger.error(f"{RED}[!] API Error fetching favorites: {e}{OFF}")
            return {}
            
    def add_favorite_album(self, album_id):
        """Adds an album to the user's Qobuz favorites."""
        return self.api_call(
            "favorite/create", 
            album_ids=str(album_id),
            artist_ids="",
            track_ids=""
        )
        
    # NEW GET_TRACK_URL (Patch 0004)
    def get_track_url(self, id, fmt_id, force_segments=False):
        # Quick fallback for MP3
        if int(fmt_id) == 5:
            return self.api_call("track/getFileUrl", id=id, fmt_id=fmt_id)

        # If not forcing segments, try the good old fast Direct URL first
        if not force_segments:
            try:
                track = self.api_call("track/getFileUrl", id=id, fmt_id=fmt_id)
                if "url" in track:
                    return track
            except Exception:
                pass # If Qobuz refuses to give the direct URL, fallback to segments automatically

        # "WEB PLAYER" METHOD (SEGMENTED DOWNLOAD)
        if self.session_id is None:
            session = self.api_call("session/start")
            self.session_id = session["session_id"]
            self.session_infos = session["infos"]
            self.session_key = self._derive_session_key()
            self.session.headers.update({"X-Session-Id": self.session_id})

        track = self.api_call("file/url", id=id, fmt_id=fmt_id)
        if "bits_depth" in track and "bit_depth" not in track:
            track["bit_depth"] = track["bits_depth"]
        if track.get("sampling_rate", 0) > 1000:
            track["sampling_rate"] = track["sampling_rate"] / 1000
        if "key" in track:
            track["raw_key"] = self._unwrap_track_key(track["key"])
        return track

    def get_artist_meta(self, id): return self.multi_meta("artist/get", "albums_count", id, None)
    def get_plist_meta(self, id): return self.multi_meta("playlist/get", "tracks_count", id, None)
    def get_label_meta(self, id): return self.multi_meta("label/get", "albums_count", id, None)
    def get_album_meta(self, id): return self.api_call("album/get", id=id)
    
    def cfg_setup(self):
        for secret in self.secrets:
            try:
                self.api_call("track/getFileUrl", id=5966783, fmt_id=5, sec=secret)
                self.sec = secret
                break
            except: continue
        if not self.sec and self.secrets: self.sec = self.secrets[0]
        if not self.sec: raise InvalidAppSecretError("No secret found.")